import networkx as nx
import numpy as np
import os.path as osp
import pickle as pkl
import scipy.sparse as sp
import torch

from sgl.data.base_data import Graph
from sgl.data.base_dataset import NodeDataset
from sgl.dataset.utils import pkl_read_file, download_to


class Planetoid(NodeDataset):
    def __init__(self, name="cora", root="./", split="official"):
        if name not in ["cora", "citeseer", "pubmed"]:
            raise ValueError("Dataset name not supported!")
        super(Planetoid, self).__init__(root + "Planetoid/", name)

        self._data = pkl_read_file(self.processed_file_paths)
        self._split = split
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(split)

    @property
    def raw_file_paths(self):
        filenames = ["x", "tx", "allx", "y", "ty", "ally", "graph", "test.index"]
        return [osp.join(self._raw_dir, "ind.{}.{}".format(self._name, filename)) for filename in filenames]

    @property
    def processed_file_paths(self):
        filename = "graph"
        return osp.join(self._processed_dir, "{}.{}".format(self._name, filename))

    def _download(self):
        url = "https://github.com/kimiyoung/planetoid/raw/master/data"
        for filepath in self.raw_file_paths:
            file_url = url + '/' + osp.basename(filepath)
            print(file_url)
            download_to(file_url, filepath)

    def _normalize(self, mx):
        """Row-normalize sparse matrix"""
        rowsum = np.array(mx.sum(1))
        r_inv = np.power(rowsum, -1).flatten()
        r_inv[np.isinf(r_inv)] = 0.
        r_mat_inv = sp.diags(r_inv)
        mx = r_mat_inv.dot(mx)
        return mx
            
    def _process(self):
        objects = []
        for raw_file in self.raw_file_paths[:-1]:
            objects.append(pkl_read_file(raw_file))

        x, tx, allx, y, ty, ally, graph = tuple(objects)

        test_idx_reorder = []
        with open(self.raw_file_paths[-1], 'r') as rf:
            try:
                for line in rf:
                    test_idx_reorder.append(int(line.strip()))
            except IOError as e:
                print(e)
                exit(1)
        test_idx_range = np.sort(test_idx_reorder)

        if self._name == "citeseer":
            # Fix citeseer dataset (there are some isolated nodes in the graph)
            # Find isolated nodes, add them as zero-vectors into the right position
            test_idx_range_full = range(
                min(test_idx_reorder), max(test_idx_reorder) + 1)
            tx_extended = sp.lil_matrix((len(test_idx_range_full), x.shape[1]))
            tx_extended[test_idx_range - min(test_idx_range), :] = tx
            tx = tx_extended
            ty_extended = np.zeros((len(test_idx_range_full), y.shape[1]))
            ty_extended[test_idx_range - min(test_idx_range), :] = ty
            ty = ty_extended

        features = sp.vstack((allx, tx)).tolil()
        features[test_idx_reorder, :] = features[test_idx_range, :]
        features = self._normalize(features)
        features = np.array(features.todense())
        num_node = features.shape[0]
        node_type = "paper"

        adj = sp.coo_matrix(nx.adjacency_matrix(nx.from_dict_of_lists(graph)))
        row, col, edge_weight = adj.row, adj.col, adj.data
        edge_type = "paper__to__paper"

        labels = np.vstack((ally, ty))
        labels[test_idx_reorder, :] = labels[test_idx_range, :]
        labels = np.argmax(labels, 1)
        labels = torch.LongTensor(labels)

        g = Graph(row, col, edge_weight, num_node, node_type, edge_type, x=features, y=labels)
        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "official":
            train_idx = range(self.num_classes * 20)
            val_idx = range(self.num_classes * 20, self.num_classes * 20 + 500)
            test_idx = range(self.num_node - 1000, self.num_node)
        elif split == "random":
            raise NotImplementedError
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



import os.path as osp
import pickle as pkl
import torch
from torch_geometric.datasets import HGBDataset
from typing import Tuple

from sgl.data.base_data import HeteroGraph
from sgl.data.base_dataset import HeteroNodeDataset
from sgl.dataset.utils import pkl_read_file


class Acm(HeteroNodeDataset):
    NODE_TYPES = [
        'paper',
        'author',
        'subject',
        'term'
    ]

    TYPE_OF_NODE_TO_PREDICT = 'paper'

    EDGE_TYPE_DELIMITER = '__to__'

    EDGE_TYPES_TUPLE = [
        ('paper', 'cite', 'paper'),
        ('paper', 'ref', 'paper'),
        ('paper', 'to', 'author'),
        ('author', 'to', 'paper'),
        ('paper', 'to', 'subject'),
        ('subject', 'to', 'paper'),
        ('paper', 'to', 'term'),
        ('term', 'to', 'paper')
    ]

    def __init__(self, root="./", split="official"):
        name = 'acm'
        # This dataset is processed based on the output of HGBDataset
        self.src_dataset = HGBDataset(root=root, name='acm').data

        super(Acm, self).__init__(root + "hgb/", name)

        self._data = pkl_read_file(self.processed_file_paths)
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(
            split)

    def edge_type_tuple_to_str(self, edge_type_tuple: Tuple) -> str:
        if len(edge_type_tuple) != 3:
            raise ValueError('number of elements is invalid for input tuple')
        return edge_type_tuple[0] + self.EDGE_TYPE_DELIMITER + edge_type_tuple[2]

    @property
    def raw_file_paths(self):
        filepath = "hgb_" + self._name + "/raw/geometric_data_processed.pt"
        return osp.join(self._raw_dir, filepath)

    @property
    def processed_file_paths(self):
        filename = "graph"
        return osp.join(self._processed_dir, "{}.{}".format(self._name, filename))

    def _download(self):
        pass

    def _process(self):

        # obtain num_node_dict
        num_node_dict = {}
        for node_type in self.NODE_TYPES:
            if 'x' in self.src_dataset[node_type]:
                num_node_dict[node_type] = self.src_dataset[node_type]['x'].size(
                    0)
            else:
                num_node_dict[node_type] = self.src_dataset[node_type]['num_nodes']
            print(f'number of {node_type} node is {num_node_dict[node_type]}')

        # obtain row_dict, col_dict, edge_weight_dict
        previous_node_cnt_dict = {}
        previous_node_cnt = 0
        for node_type in self.NODE_TYPES:
            previous_node_cnt_dict[node_type] = previous_node_cnt
            previous_node_cnt += num_node_dict[node_type]

        row_dict = {}
        col_dict = {}
        edge_weight_dict = {}
        for edge_type_tuple in self.EDGE_TYPES_TUPLE:
            edge_type = self.edge_type_tuple_to_str(edge_type_tuple)

            row_dict[edge_type] = self.src_dataset[edge_type_tuple]['edge_index'][0] \
                                  + previous_node_cnt_dict[edge_type_tuple[0]]

            col_dict[edge_type] = self.src_dataset[edge_type_tuple]['edge_index'][1] \
                                  + previous_node_cnt_dict[edge_type_tuple[2]]

            edge_weight_dict[edge_type] = torch.ones(
                self.src_dataset[edge_type_tuple]['edge_index'].size(1))

        # obtain node_types
        node_types = self.NODE_TYPES

        # obtain edge_types_found
        edge_types_found = [self.edge_type_tuple_to_str(
            edge_type_tuple) for edge_type_tuple in self.EDGE_TYPES_TUPLE]

        # obtain node_id_dict
        node_id_dict = {}
        accumulated_node_cnt = 0
        for node_type in self.NODE_TYPES:
            num_cur_node_type = num_node_dict[node_type]
            node_id_dict[node_type] = [i for i in range(
                accumulated_node_cnt,
                accumulated_node_cnt + num_cur_node_type)]
            accumulated_node_cnt += num_cur_node_type

        # obtain x_dict
        x_dict = {}

        total_dim = 0

        start_pos_of_node_type = {}
        for node_type in self.NODE_TYPES:
            start_pos_of_node_type[node_type] = total_dim
            if 'x' in self.src_dataset[node_type]:
                total_dim += self.src_dataset[node_type]['x'].size(1)

        accumulated_feature_dim = 0
        for node_type in self.NODE_TYPES:
            padded_tensor = torch.zeros(num_node_dict[node_type], total_dim)
            if 'x' in self.src_dataset[node_type]:
                cur_x_len = self.src_dataset[node_type]['x'].size(1)
                padded_tensor[:, accumulated_feature_dim:accumulated_feature_dim +
                                                         cur_x_len] = self.src_dataset[node_type]['x']
                accumulated_feature_dim += cur_x_len
            x_dict[node_type] = padded_tensor.numpy()

        # obtain y_dict
        y_dict = {}

        for node_type in self.NODE_TYPES:
            if 'y' in self.src_dataset[node_type]:
                y_dict[node_type] = self.src_dataset[node_type]['y']
            else:
                y_dict[node_type] = None

        g = HeteroGraph(row_dict, col_dict, edge_weight_dict, num_node_dict, node_types, edge_types_found, node_id_dict,
                        x_dict, y_dict)

        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "official":
            train_idx = self.src_dataset[self.TYPE_OF_NODE_TO_PREDICT]['train_mask'].nonzero(
            ).flatten()
            test_idx = self.src_dataset[self.TYPE_OF_NODE_TO_PREDICT]['test_mask'].nonzero(
            ).flatten()
        elif split == "random":
            raise NotImplementedError
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, None, test_idx



import time
import torch
import torch.nn as nn
from torch.optim import Adam
from torch.utils.data import DataLoader
import torch.nn.functional as F
import numpy as np

from sgl.tasks.base_task import BaseTask
from sgl.tasks.utils import accuracy, set_seed, train, mini_batch_train, \
                            evaluate, mini_batch_evaluate, add_labels

# Node Classification with Label use and Label Reuse trick
# NOTE: When use this trick, the input_dim of model should be feature_dim + num_classes instead of feature_dim
class NodeClassificationWithLabelUse(BaseTask):
    def __init__(self, dataset, model, lr, weight_decay, epochs, device, loss_fn=nn.CrossEntropyLoss(), seed=42,
                 train_batch_size=None, eval_batch_size=None, label_reuse_batch_size=None,
                 mask_rate=0.5, use_labels=True, reuse_start_epoch=0, label_iters=0):
        super(NodeClassificationWithLabelUse, self).__init__()

        self.__dataset = dataset
        self.__labels = self.__dataset.y

        self.__model = model
        self.__optimizer = Adam(model.parameters(), lr=lr,
                                weight_decay=weight_decay)
        self.__epochs = epochs
        self.__loss_fn = loss_fn
        self.__device = device
        self.__seed = seed
        self.__mask_rate = mask_rate
        self.__use_labels = use_labels
        self.__reuse_start_epoch = reuse_start_epoch
        self.__label_iters = label_iters
        self.__label_reuse_batch_size = label_reuse_batch_size

        self.__mini_batch = False
        if train_batch_size is not None:
            self.__mini_batch = True
            self.__train_loader = DataLoader(
                self.__dataset.train_idx, batch_size=train_batch_size, shuffle=True, drop_last=False)
            self.__val_loader = DataLoader(
                self.__dataset.val_idx, batch_size=eval_batch_size, shuffle=False, drop_last=False)
            self.__test_loader = DataLoader(
                self.__dataset.test_idx, batch_size=eval_batch_size, shuffle=False, drop_last=False)
            self.__all_eval_loader = DataLoader(
                range(self.__dataset.data.num_node), batch_size=eval_batch_size, shuffle=False, drop_last=False)

        if self.__label_iters > 0 and self.__use_labels is False:
            raise ValueError("When using label reuse, it's essential to enable label use!")

        self.__test_acc = self._execute()

    @property
    def test_acc(self):
        return self.__test_acc

    def _execute(self):
        set_seed(self.__seed)

        self.__model = self.__model.to(self.__device)
        self.__labels = self.__labels.to(self.__device)

        t_total = time.time()
        best_val = 0.
        best_test = 0.

        features = self.__dataset.x
        for epoch in range(self.__epochs):
            # add label feature if use_label is True
            if self.__use_labels:
                train_idx = np.array(self.__dataset.train_idx)
                mask = np.random.rand(train_idx.shape[0]) < self.__mask_rate
                train_labels_idx = train_idx[mask]
                train_pred_idx = train_idx[~mask]
                features = add_labels(self.__dataset.x, self.__dataset.y, train_labels_idx, self.__dataset.num_classes)
            
            pre_time_st = time.time()
            self.__model.preprocess(self.__dataset.adj, features)
            pre_time_ed = time.time()
            print(f"Feature Propagate done in {(pre_time_ed - pre_time_st):.4f}s")

            # label reuse
            # small optimization: only utilize the predicted soft_labels in later epoches
            if self.__label_iters > 0 and epoch > self.__reuse_start_epoch:
                model = self.__model
                full_idx = torch.arange(self.__dataset.num_node, dtype=torch.long)
                val_idx = np.array(self.__dataset.val_idx)
                test_idx = np.array(self.__dataset.test_idx)
                unlabeled_idx = np.concatenate([train_pred_idx, val_idx, test_idx])
                for _ in range(self.__label_iters):
                    pred = []
                    if self.__label_reuse_batch_size is not None:
                        label_reuse_loader = DataLoader(full_idx, batch_size=self.__label_reuse_batch_size, shuffle=False, drop_last=False)
                        for batch in label_reuse_loader:
                            tmp = model.model_forward(batch, self.__device)
                            pred.append(tmp)
                        pred = torch.cat(pred)
                    else:
                        pred = model.model_forward(full_idx, self.__device)
                    pred = pred.detach().cpu()
                    torch.cuda.empty_cache()
                    features[unlabeled_idx, -self.__dataset.num_classes:] = F.softmax(pred[unlabeled_idx], dim=-1)
                    self.__model.preprocess(self.__dataset.adj, features)

            t = time.time()
            if self.__mini_batch is False:
                loss_train, acc_train = train(self.__model, train_pred_idx, self.__labels, self.__device,
                                              self.__optimizer, self.__loss_fn)
                acc_val, acc_test = evaluate(self.__model, self.__dataset.val_idx, self.__dataset.test_idx,
                                             self.__labels, self.__device)
            else:
                loss_train, acc_train = mini_batch_train(self.__model, train_pred_idx, self.__train_loader,
                                                         self.__labels, self.__device, self.__optimizer, self.__loss_fn)
                acc_val, acc_test = mini_batch_evaluate(self.__model, self.__dataset.val_idx, self.__val_loader,
                                                        self.__dataset.test_idx, self.__test_loader, self.__labels,
                                                        self.__device)

            print('Epoch: {:03d}'.format(epoch + 1),
                  'loss_train: {:.4f}'.format(loss_train),
                  'acc_train: {:.4f}'.format(acc_train),
                  'acc_val: {:.4f}'.format(acc_val),
                  'acc_test: {:.4f}'.format(acc_test),
                  'time: {:.4f}s'.format(time.time() - t))
            if acc_val > best_val:
                best_val = acc_val
                best_test = acc_test

        acc_val, acc_test = self._postprocess()
        if acc_val > best_val:
            best_val = acc_val
            best_test = acc_test

        print("Optimization Finished!")
        print("Total time elapsed: {:.4f}s".format(time.time() - t_total))
        print(f'Best val: {best_val:.4f}, best test: {best_test:.4f}')
        return best_test

    def _postprocess(self):
        self.__model.eval()
        if self.__mini_batch is False:
            outputs = self.__model.model_forward(
                range(self.__dataset.num_node), self.__device).to("cpu")
        else:
            outputs = None
            for batch in self.__all_eval_loader:
                output = self.__model.model_forward(batch, self.__device)
                if outputs is None:
                    outputs = output
                else:
                    outputs = torch.vstack((outputs, output))

        final_output = self.__model.postprocess(self.__dataset.adj, outputs)
        acc_val = accuracy(
            final_output[self.__dataset.val_idx], self.__labels[self.__dataset.val_idx])
        acc_test = accuracy(
            final_output[self.__dataset.test_idx], self.__labels[self.__dataset.test_idx])
        return acc_val, acc_test



from typing import List


def GetGpuMemory() -> List[int]:
    import os
    import random

    tmp_file_name = f'gpu_mem_list_{random.randint(0, 10000000)}.txt'
    while os.path.exists(tmp_file_name):
        tmp_file_name = f'gpu_mem_list_{random.randint(0, 10000000)}.txt'

    os.system(
        f'nvidia-smi -q -d Memory | grep -A4 GPU | grep Free > {tmp_file_name}')
    gpu_mem_list = [int(x.split()[2])
                    for x in open(tmp_file_name, 'r').readlines()]
    os.system(f'rm {tmp_file_name}')
    return gpu_mem_list


def IdxOfMaxElement(list: List[int]) -> int:
    return list.index(max(list))


# This is is the only API you should call from outside
# 'display' determines if the chosen result will be printed
def GpuWithMaxFreeMem(display: bool = True) -> int:
    result = IdxOfMaxElement(GetGpuMemory())
    if display:
        print(f'GPU {result} is chosen')
    return result



import os
import time
import torch
import torch.nn as nn
from torch.optim import Adam
import torch.distributed as dist
import torch.multiprocessing as mp
from torch.utils.data import Dataset, DataLoader

from sgl.search.base_search import BaseSearch
from sgl.search.utils import accuracy, set_seed

class TrainDataset(Dataset):
    def __init__(self, data, label):
        self.data = data
        self.label = label

    def __len__(self):
        return len(self.data)

    def __getitem__(self, ind):
        x = self.data[ind]
        y = self.label[ind]
        return x, y

class SearchManagerDist(BaseSearch):
    def __init__(self, dataset, model, seed=42):
        super(SearchManagerDist, self).__init__()

        self.__dataset = dataset
        self.__model = model
        self.__seed = seed
    
    def _execute(self, args):
        t_pre_start = time.time()
        self.__model.preprocess(self.__dataset.adj, self.__dataset.x)
        t_pre_end = time.time()
        time_preprocess = t_pre_end - t_pre_start

        args.world_size = args.gpus * args.nodes
        os.environ['MASTER_ADDR'] = '127.0.0.1'
        os.environ['MASTER_PORT'] = '1788'
        
        t_total = time.time()
        best_val = 0.
        best_test = 0.
        for i in range(10):
            mp.spawn(self._train, nprocs=args.gpus, args=(args, self.__dataset, self.__model))
            acc_val, acc_test = self._evaluate()
            print('Turn: {:03d}'.format(i + 1),
                    'acc_val: {:.4f}'.format(acc_val),
                    'acc_test: {:.4f}'.format(acc_test))
            if acc_val > best_val:
                best_val = acc_val
                best_test = acc_test
                torch.save(self.__model, './best.pt')

        acc_val, acc_test, time_forward = self._postprocess()
        print(f'Post val: {acc_val:.4f}, Post test: {acc_test:.4f}')
        if acc_val > best_val:
            best_val = acc_val
            best_test = acc_test
        total_time = time_preprocess + time_forward

        print("Optimization Finished!")
        print("Total time elapsed: {:.4f}s".format(time.time() - t_total))
        print(f'Best val: {best_val:.4f}, best test: {best_test:.4f}')
        return best_test, total_time

    def _train(self, gpu, args, dataset, model):
        rank = args.nr * args.gpus + gpu
        dist.init_process_group(backend='nccl', init_method='env://', world_size=args.world_size, rank=rank)
        set_seed(self.__seed)

        torch.cuda.set_device(gpu)
        model.cuda(gpu)
        criterion = nn.CrossEntropyLoss().cuda(gpu)
        optimizer = Adam(model.parameters(), lr=args.lr, weight_decay=args.wd)

        _processed_feat_list = model._processed_feat_list
        model = nn.parallel.DistributedDataParallel(model, device_ids=[gpu])

        train_dataset = TrainDataset(self.__dataset.train_idx, self.__dataset.y[self.__dataset.train_idx])
        train_sampler = torch.utils.data.distributed.DistributedSampler(train_dataset,
                                                                        num_replicas=args.world_size,
                                                                        rank=rank)
        train_loader = torch.utils.data.DataLoader(dataset=train_dataset,
                                                batch_size=args.batch,
                                                shuffle=False,
                                                num_workers=0,
                                                pin_memory=True,
                                                sampler=train_sampler)
        
        for epoch in range(args.epochs):
            for i, (idx, labels) in enumerate(train_loader):
                transferred_feat_list = [feat[idx].cuda(
                    non_blocking=True) for feat in _processed_feat_list]
                labels = labels.cuda(
                    non_blocking=True)
                # Forward pass
                outputs = model(transferred_feat_list)
                loss = criterion(outputs, labels)

                # Backward and optimize
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

    def _postprocess(self):
        device = torch.device(f"cuda:{0}" if torch.cuda.is_available() else "cpu")
        model = torch.load('./best.pt')
        model = model.to(device)
        model.eval()

        t_forward_start = time.time()
        output = model.model_forward(range(self.__dataset.num_node), device)
        final_output = model.postprocess(self.__dataset.adj, output)
        t_forward_end = time.time()
        time_forward = t_forward_end - t_forward_start

        acc_val = accuracy(final_output[self.__dataset.val_idx], self.__dataset.y[self.__dataset.val_idx])
        acc_test = accuracy(final_output[self.__dataset.test_idx], self.__dataset.y[self.__dataset.test_idx])
        return acc_val, acc_test, time_forward

    def _evaluate(self):
        device = torch.device(f"cuda:{0}" if torch.cuda.is_available() else "cpu")
        self.__model = self.__model.to(device)
        self.__model.eval()

        val_output = self.__model.model_forward(self.__dataset.val_idx, device)
        test_output = self.__model.model_forward(self.__dataset.test_idx, device)

        acc_val = accuracy(val_output, self.__dataset.y[self.__dataset.val_idx])
        acc_test = accuracy(test_output, self.__dataset.y[self.__dataset.test_idx])
        return acc_val, acc_test



from sgl.operators.base_op import MessageOp


class LastMessageOp(MessageOp):
    def __init__(self):
        super(LastMessageOp, self).__init__()
        self._aggr_type = "last"

    def _combine(self, feat_list):
        return feat_list[-1]



import numpy as np
import os.path as osp
import pickle as pkl
import torch
import torch.nn.functional as F
from scipy.io import loadmat

from sgl.data.base_data import Graph
from sgl.data.base_dataset import NodeDataset
from sgl.dataset.utils import pkl_read_file, download_to


# A variety of non-homophilous graph datasets
# Note that only "penn94" has official split_mask, split_id in [0, 9] to identify different 
# For other datasets applied official split, you should set numbers of samples for training/valid set
class LINKXDataset(NodeDataset):
    def __init__(self, name="penn94", root="./", split="official", split_id=0, num_train_per_class=10,
                 num_valid_per_class=10):
        name = name.lower()
        if name not in ['penn94', 'reed98', 'amherst41', 'cornell5', 'johnshopkins55']:
            raise ValueError("Dataset name not supported!")
        if name in ['penn94'] and split_id not in range(5):
            raise ValueError("Split id not supported!")
        super(LINKXDataset, self).__init__(root + 'LINKXDataset', name)

        self._data = pkl_read_file(self.processed_file_paths)
        self._split, self._split_id = split, split_id
        self._num_train_per_class = num_train_per_class
        self._num_valid_per_class = num_valid_per_class
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(
            split)

    @property
    def raw_file_paths(self):
        dataset_name = {
            'penn94': 'Penn94.mat',
            'reed98': 'Reed98.mat',
            'amherst41': 'Amherst41.mat',
            'cornell5': 'Cornell5.mat',
            'johnshopkins55': 'Johns%20Hopkins55.mat',
        }
        splits_name = {
            'penn94': 'fb100-Penn94-splits.npy',
        }
        filenames = [dataset_name[self._name]]
        if self._name in splits_name:
            filenames += [splits_name[self._name]]

        return [osp.join(self._raw_dir, filename) for filename in filenames]

    @property
    def processed_file_paths(self):
        return osp.join(self._processed_dir, f"{self._name}.graph")

    def _download(self):
        url = 'https://github.com/CUAI/Non-Homophily-Large-Scale/raw/master/data'

        datasets = {
            'penn94': f'{url}/facebook100/Penn94.mat',
            'reed98': f'{url}/facebook100/Reed98.mat',
            'amherst41': f'{url}/facebook100/Amherst41.mat',
            'cornell5': f'{url}/facebook100/Cornell5.mat',
            'johnshopkins55': f'{url}/facebook100/Johns%20Hopkins55.mat',
        }

        splits = {
            'penn94': f'{url}/splits/fb100-Penn94-splits.npy',
        }

        file_url = datasets[self._name]
        print(file_url)
        download_to(file_url, self.raw_file_paths[0])

        if self._name in splits:
            file_url = splits[self._name]
            print(file_url)
            download_to(file_url, self.raw_file_paths[1])

    def _process(self):
        mat = loadmat(self.raw_file_paths[0])

        edge_index = mat['A'].tocsr().tocoo()
        row, col = edge_index.row, edge_index.col
        edge_weight = torch.ones(len(row))
        edge_type = "user__to__user"

        metadata = torch.from_numpy(mat['local_info'].astype('int64'))

        xs = []
        labels = metadata[:, 1] - 1  # gender label, -1 means unlabeled
        x = torch.cat([metadata[:, :1], metadata[:, 2:]], dim=-1)
        for i in range(x.size(1)):
            _, out = x[:, i].unique(return_inverse=True)
            xs.append(F.one_hot(out).to(torch.float))
        x = torch.cat(xs, dim=-1)

        features = x.numpy()
        num_node = features.shape[0]
        node_type = "user"

        g = Graph(row, col, edge_weight, num_node,
                  node_type, edge_type, x=features, y=labels)

        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        have_split = ['penn94']
        # datasets that have official split_mask
        if self._name in have_split:
            if split == 'official':
                split_full = np.load(self.raw_file_paths[1], allow_pickle=True)

                split_idx = split_full[self._split_id]
                train_idx, val_idx, test_idx = split_idx['train'], split_idx['valid'], split_idx['test']

            else:
                raise ValueError("Please input valid split pattern!")

        else:
            if split == 'official':

                mat = loadmat(self.raw_file_paths[0])
                metadata = mat['local_info'].astype('int64')
                labels = metadata[:, 1] - 1  # gender label, -1 means unlabeled

                num_train_per_class = self._num_train_per_class
                num_val = self._num_valid_per_class
                train_idx, val_idx, test_idx = np.empty(0), np.empty(0), np.empty(0)
                for i in range(self.num_classes):
                    idx = np.nonzero(labels == i)[0]
                    train_idx = np.append(train_idx, idx[:num_train_per_class])
                    val_idx = np.append(val_idx, idx[num_train_per_class: num_train_per_class + num_val])
                    test_idx = np.append(test_idx, idx[num_train_per_class + num_val:])
                train_idx.reshape(-1)
                val_idx.reshape(-1)
                test_idx.reshape(-1)

            else:
                raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



import random
import math
import torch
import torch.nn.functional as F
import numpy as np
import scipy.sparse as sp
from sklearn.cluster import KMeans
from sklearn.metrics import roc_auc_score, average_precision_score

from sgl.tasks.clustering_metrics import clustering_metrics

def accuracy(output, labels):
    pred = output.max(1)[1].type_as(labels)
    correct = pred.eq(labels).double()
    correct = correct.sum()
    return (correct / len(labels)).item()


def set_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

def adjust_learning_rate(optimizer, lr, epoch):
    if epoch <= 50:
        for param_group in optimizer.param_groups:
            param_group["lr"] = lr * epoch / 50

def add_labels(features, labels, idx, num_classes):
    onehot = np.zeros([features.shape[0], num_classes])
    onehot[idx, labels[idx]] = 1
    return np.concatenate([features, onehot], axis=-1)

def evaluate(model, val_idx, test_idx, labels, device):
    model.eval()
    val_output = model.model_forward(val_idx, device)
    test_output = model.model_forward(test_idx, device)

    acc_val = accuracy(val_output, labels[val_idx])
    acc_test = accuracy(test_output, labels[test_idx])
    return acc_val, acc_test


def mini_batch_evaluate(model, val_idx, val_loader, test_idx, test_loader, labels, device):
    model.eval()
    correct_num_val, correct_num_test = 0, 0
    for batch in val_loader:
        val_output = model.model_forward(batch, device)
        pred = val_output.max(1)[1].type_as(labels)
        correct_num_val += pred.eq(labels[batch]).double().sum()
    acc_val = correct_num_val / len(val_idx)

    for batch in test_loader:
        test_output = model.model_forward(batch, device)
        pred = test_output.max(1)[1].type_as(labels)
        correct_num_test += pred.eq(labels[batch]).double().sum()
    acc_test = correct_num_test / len(test_idx)

    return acc_val.item(), acc_test.item()


def train(model, train_idx, labels, device, optimizer, loss_fn):
    model.train()
    optimizer.zero_grad()

    train_output = model.model_forward(train_idx, device)
    loss_train = loss_fn(train_output, labels[train_idx])
    acc_train = accuracy(train_output, labels[train_idx])
    loss_train.backward()
    optimizer.step()

    return loss_train.item(), acc_train


def mini_batch_train(model, train_idx, train_loader, labels, device, optimizer, loss_fn):
    model.train()
    correct_num = 0
    loss_train_sum = 0.
    for batch in train_loader:
        train_output = model.model_forward(batch, device)
        loss_train = loss_fn(train_output, labels[batch])

        pred = train_output.max(1)[1].type_as(labels)
        correct_num += pred.eq(labels[batch]).double().sum()
        loss_train_sum += loss_train.item()

        optimizer.zero_grad()
        loss_train.backward()
        optimizer.step()

    loss_train = loss_train_sum / len(train_loader)
    acc_train = correct_num / len(train_idx)

    return loss_train, acc_train.item()


def cluster_loss(train_output, y_pred, cluster_centers):

    for i in range(len(cluster_centers)):
        if i == 0:
            dist = torch.norm(train_output - cluster_centers[i], p=2, dim=1, keepdim=True)
        else:
            dist = torch.cat((dist, torch.norm(train_output - cluster_centers[i], p=2, dim=1, keepdim=True)), 1)
    
    loss = 0.
    loss_tmp = -dist.mean(1).sum()
    loss_tmp += 2 * np.sum(dist[j, x] for j, x in zip(range(dist.shape[0]), y_pred))
    loss = loss_tmp / dist.shape[0]
    return loss


def clustering_train(model, train_idx, labels, device, optimizer, loss_fn, n_clusters, n_init):
    model.train()
    optimizer.zero_grad()

    train_output = model.model_forward(train_idx, device)
    
    # calc loss
    kmeans = KMeans(n_clusters=n_clusters, n_init=n_init)
    y_pred = kmeans.fit_predict(train_output.data.cpu().numpy()) # cluster_label
    cluster_centers = torch.FloatTensor(kmeans.cluster_centers_).to(device)

    loss_train = loss_fn(train_output, y_pred, cluster_centers)
    loss_train.backward()
    optimizer.step()

    # calc acc, nmi, adj
    labels = labels.cpu().numpy()
    cm = clustering_metrics(labels, y_pred)
    acc, nmi, adjscore = cm.evaluationClusterModelFromLabel()

    return loss_train.item(), acc, nmi, adjscore


def sparse_to_tuple(sparse_mx):
    if not sp.isspmatrix_coo(sparse_mx):
        sparse_mx = sparse_mx.tocoo()
    coords = np.vstack((sparse_mx.row, sparse_mx.col)).transpose()
    values = sparse_mx.data
    shape = sparse_mx.shape
    return coords, values, shape


def mask_test_edges(adj):
    # Function to build test set with 10% positive links
    # NOTE: Splits are randomized and results might slightly deviate from reported numbers in the paper.
    # TODO: Clean up.

    # Remove diagonal elements
    adj = adj - sp.dia_matrix((adj.diagonal()[np.newaxis, :], [0]), shape=adj.shape)
    adj.eliminate_zeros()
    # Check that diag is zero:
    assert np.diag(adj.todense()).sum() == 0

    adj_triu = sp.triu(adj)
    adj_tuple = sparse_to_tuple(adj_triu)
    edges = adj_tuple[0]
    edges_all = sparse_to_tuple(adj)[0]
    num_test = int(np.floor(edges.shape[0] / 10.))
    num_val = int(np.floor(edges.shape[0] / 20.))

    edges_all_set = set()
    for edge in edges_all.tolist():
        edges_all_set.add(tuple(edge))

    all_edge_idx = list(range(edges.shape[0]))
    np.random.shuffle(all_edge_idx)
    val_edge_idx = all_edge_idx[:num_val]
    test_edge_idx = all_edge_idx[num_val:(num_val + num_test)]
    test_edges = edges[test_edge_idx]
    val_edges = edges[val_edge_idx]
    train_edges = np.delete(edges, np.hstack([test_edge_idx, val_edge_idx]), axis=0)

    def ismember(a, b):
        return a in b

    train_edges_false = set()
    while len(train_edges_false) < len(train_edges):
        idx_i = np.random.randint(0, adj.shape[0])
        idx_j = np.random.randint(0, adj.shape[0])
        if idx_i == idx_j:
            continue
        if ismember( (idx_i, idx_j), edges_all_set):
            continue
        if train_edges_false:
            if ismember( (idx_j, idx_i), train_edges_false):
                continue
            if ismember( (idx_i, idx_j), train_edges_false):
                continue
        train_edges_false.add((idx_i, idx_j))

    test_edges_false = set()
    while len(test_edges_false) < len(test_edges):
        idx_i = np.random.randint(0, adj.shape[0])
        idx_j = np.random.randint(0, adj.shape[0])
        if idx_i == idx_j:
            continue
        if ismember( (idx_i, idx_j), edges_all_set):
            continue
        if ismember( (idx_i, idx_j), train_edges_false):
            continue
        if ismember( (idx_j, idx_i), train_edges_false):
            continue
        if test_edges_false:
            if ismember( (idx_j, idx_i), test_edges_false):
                continue
            if ismember( (idx_i, idx_j), test_edges_false):
                continue
        test_edges_false.add((idx_i, idx_j))

    val_edges_false = set()
    while len(val_edges_false) < len(val_edges):
        idx_i = np.random.randint(0, adj.shape[0])
        idx_j = np.random.randint(0, adj.shape[0])
        if idx_i == idx_j:
            continue
        if ismember( (idx_i, idx_j), edges_all_set):
            continue
        if ismember( (idx_i, idx_j), train_edges_false):
            continue
        if ismember( (idx_j, idx_i), train_edges_false):
            continue
        if ismember( (idx_i, idx_j), test_edges_false):
            continue
        if ismember( (idx_j, idx_i), test_edges_false):
            continue
        if val_edges_false:
            if ismember( (idx_j, idx_i), np.array(val_edges_false)):
                continue
            if ismember( (idx_i, idx_j), np.array(val_edges_false)):
                continue
        val_edges_false.add((idx_i, idx_j))

    #assert ~ismember(test_edges_false, edges_all)
    #assert ~ismember(val_edges_false, edges_all)
    #assert ~ismember(val_edges, train_edges)
    #assert ~ismember(test_edges, train_edges)
    #assert ~ismember(val_edges, test_edges)

    data = np.ones(train_edges.shape[0])

    # Re-build adj matrix
    adj_train = sp.csr_matrix((data, (train_edges[:, 0], train_edges[:, 1])), shape=adj.shape)
    adj_train = adj_train + adj_train.T

    train_edges_false = list(train_edges_false)
    val_edges_false = list(val_edges_false)
    test_edges_false = list(test_edges_false)

    train_edges, train_edges_false  = torch.tensor(train_edges), torch.tensor(train_edges_false)
    val_edges, val_edges_false = torch.tensor(val_edges), torch.tensor(val_edges_false)
    test_edges, test_edges_false = torch.tensor(test_edges), torch.tensor(test_edges_false)

    # NOTE: these edge lists only contain single direction of edge!
    return adj_train, train_edges, train_edges_false, val_edges, val_edges_false, test_edges, test_edges_false


# input full edge_features, pos_edges and neg_edges to calc roc_auc, avg_prec score
def edge_predict_score(edge_feature, pos_edges, neg_edges, threshold):
    labels = torch.cat((torch.ones(len(pos_edges)), torch.zeros(len(neg_edges))))
    all_edges = torch.cat((pos_edges, neg_edges))
    edge_pred = edge_feature[all_edges[:, 0], all_edges[:, 1]].reshape(-1)
    edge_pred = torch.sigmoid(edge_pred)
    edge_pred = edge_pred > threshold
    roc_auc = roc_auc_score(labels, edge_pred)
    avg_prec = average_precision_score(labels, edge_pred)
    return roc_auc, avg_prec


def edge_predict_train(model, train_node_index, with_params, pos_edges, neg_edges, 
                       device, optimizer, loss_fn, threshold):
    if with_params is True:
        model.train()
        optimizer.zero_grad()

    train_output = model.model_forward(train_node_index, device)
    edge_feature = torch.mm(train_output, train_output.t())
    labels = torch.cat((torch.ones(len(pos_edges)), torch.zeros(len(neg_edges)))).to(device)
    train_edge = torch.cat((pos_edges, neg_edges)).to(device)
    edge_pred = edge_feature[train_edge[:, 0], train_edge[:, 1]].reshape(-1)
    edge_pred = torch.sigmoid(edge_pred)

    # print("-----------------------------")
    # print("edge_features:  ", edge_feature[:200])
    # print("edge_pred:\n", edge_pred[len(pos_edges)-50:len(pos_edges)+50])
    # print("labels:\n",labels[len(pos_edges)-50:len(pos_edges)+50])
    # print("-----------------------------")

    loss = loss_fn(edge_pred, labels)
    if with_params is True:
        loss.backward()
        optimizer.step()

    labels = labels.cpu().data
    edge_pred = edge_pred.cpu().data
    edge_pred = edge_pred > threshold
    roc_auc = roc_auc_score(labels, edge_pred)
    avg_prec = average_precision_score(labels, edge_pred)
    return loss.item(), roc_auc, avg_prec


def edge_predict_eval(model, train_node_index, val_pos_edges, val_neg_edges, 
                      test_pos_edges, test_neg_edges, device, threshold):
    model.eval()
    train_output = model.model_forward(train_node_index, device)
    edge_feature = torch.mm(train_output, train_output.t()).cpu().data

    roc_auc_val, avg_prec_val = edge_predict_score(edge_feature, val_pos_edges, val_neg_edges, threshold)
    roc_auc_test, avg_prec_test = edge_predict_score(edge_feature, test_pos_edges, test_neg_edges, threshold)

    return roc_auc_val, avg_prec_val, roc_auc_test, avg_prec_test


def mini_batch_edge_predict_train(model, train_node_index, with_params, train_loader, 
                                  device, optimizer, loss_fn, threshold):
    if with_params is True:
        model.train()
        optimizer.zero_grad()
    
    loss_train = 0.
    roc_auc_sum = 0.
    avg_prec_sum = 0.

    output = model.model_forward(train_node_index, device)
    output = output.cpu()
    edge_feature = torch.mm(output, output.t())
    edge_feature = torch.sigmoid(edge_feature)

    for batch, label in train_loader:
        edge_pred = edge_feature[batch[:, 0], batch[:, 1]].reshape(-1)
        # print("-----------------------------")
        # print("edge_pred:\n", edge_pred.data[:100])
        # print("labels:\n",label.data[:100])
        # print("roc_auc_partial: ",roc_auc_score(label.data, edge_pred.data[:100]))
        # print("-----------------------------")
        pred_label = edge_pred > threshold
        roc_auc_sum += roc_auc_score(label.data, pred_label.data)
        avg_prec_sum += average_precision_score(label.data, pred_label.data)

        edge_pred = edge_pred.to(device)
        label = label.to(device)
        loss_train += loss_fn(edge_pred, label)

    if with_params is True:
        loss_train.backward()
        optimizer.step()
        
    loss_train = loss_train.item() / len(train_loader)
    roc_auc = roc_auc_sum / len(train_loader)
    avg_prec = avg_prec_sum / len(train_loader)

    return loss_train, roc_auc, avg_prec


def mini_batch_edge_predict_eval(model, train_node_index, val_loader, test_loader, device, threshold):
    model.eval()
    roc_auc_val_sum, avg_prec_val_sum = 0., 0.
    roc_auc_test_sum, avg_prec_test_sum = 0., 0.

    output = model.model_forward(train_node_index, device)
    output = output.cpu().data
    edge_feature = torch.mm(output, output.t())
    edge_feature = torch.sigmoid(edge_feature)

    for batch, label in val_loader:
        edge_pred = edge_feature[batch[:, 0], batch[:, 1]].reshape(-1)
        label_pred = edge_pred > threshold
        roc_auc_val_sum += roc_auc_score(label, label_pred)
        avg_prec_val_sum += average_precision_score(label, label_pred)

    roc_auc_val = roc_auc_val_sum / len(val_loader)
    avg_prec_val = avg_prec_val_sum / len(val_loader)

    for batch, label in test_loader:
        edge_pred = edge_feature[batch[:, 0], batch[:, 1]].reshape(-1)
        label_pred = edge_pred > threshold
        roc_auc_test_sum += roc_auc_score(label, edge_pred)
        avg_prec_test_sum += average_precision_score(label, edge_pred)

    roc_auc_test = roc_auc_test_sum / len(test_loader)
    avg_prec_test = avg_prec_test_sum / len(test_loader)

    return roc_auc_val, avg_prec_val, roc_auc_test, avg_prec_test


def mix_pos_neg_edges(pos_edges, neg_edges, mix_size):
    start, end = 0, mix_size
    mix_edges = torch.cat((pos_edges[start:end], neg_edges[start:end]))
    mix_labels = torch.cat((torch.ones(end - start), torch.zeros(end - start)))

    start += mix_size
    end += mix_size
    while end < len(pos_edges):
        tmp_edges = torch.cat((pos_edges[start:end], neg_edges[start:end]))
        tmp_labels = torch.cat((torch.ones(end - start), torch.zeros(end - start)))
        mix_edges = torch.cat((mix_edges, tmp_edges))
        mix_labels = torch.cat((mix_labels, tmp_labels))
        start += mix_size
        end += mix_size
    
    tmp_edges = torch.cat((pos_edges[start:], neg_edges[start:]))
    tmp_labels = torch.cat((torch.ones(len(pos_edges) - start), torch.zeros(len(neg_edges) - start)))
    mix_edges = torch.cat((mix_edges, tmp_edges))
    mix_labels = torch.cat((mix_labels, tmp_labels))

    return mix_edges, mix_labels

def adj_to_symmetric_norm(adj, r):
    adj = adj + sp.eye(adj.shape[0])
    degrees = np.array(adj.sum(1))
    r_inv_sqrt_left = np.power(degrees, r - 1).flatten()
    r_inv_sqrt_left[np.isinf(r_inv_sqrt_left)] = 0.
    r_mat_inv_sqrt_left = sp.diags(r_inv_sqrt_left)

    r_inv_sqrt_right = np.power(degrees, -r).flatten()
    r_inv_sqrt_right[np.isinf(r_inv_sqrt_right)] = 0.
    r_mat_inv_sqrt_right = sp.diags(r_inv_sqrt_right)

    adj_normalized = adj.dot(r_mat_inv_sqrt_left).transpose().dot(r_mat_inv_sqrt_right)
    return adj_normalized
    
def sparse_mx_to_torch_sparse_tensor(sparse_mx):
    """Convert a scipy sparse matrix to a torch sparse tensor."""
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    indices = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)
    return torch.sparse.FloatTensor(indices, values, shape)



import networkx as nx
import numpy as np
import os.path as osp
import pickle as pkl
import torch

from sgl.data.base_data import Graph
from sgl.data.base_dataset import NodeDataset
from sgl.dataset.utils import pkl_read_file


class KarateClub(NodeDataset):
    def __init__(self, name="karateclub", root="./", split="official", num_train_per_class=1, num_valid_per_class=1):
        super(KarateClub, self).__init__(root + "KarateClub/", name)

        self._data = pkl_read_file(self.processed_file_paths)
        self._split = split
        self._num_train_per_class = num_train_per_class
        self._num_valid_per_class = num_valid_per_class
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(
            split)

    @property
    def raw_file_paths(self):
        return [osp.join(self._raw_dir, "karateclub")]

    @property
    def processed_file_paths(self):
        return osp.join(self._processed_dir, f"{self._name}.graph")

    def _download(self):
        pass

    def _process(self):
        G = nx.karate_club_graph()
        features = np.eye(G.number_of_nodes())
        num_node = features.shape[0]
        node_type = "person"

        if hasattr(nx, 'to_scipy_sparse_array'):
            adj = nx.to_scipy_sparse_array(G).tocoo()
        else:
            adj = nx.to_scipy_sparse_matrix(G).tocoo()

        row, col, edge_weight = adj.row, adj.col, adj.data
        edge_type = "person__to__person"

        # Create communities
        labels = torch.tensor([
            1, 1, 1, 1, 3, 3, 3, 1, 0, 1, 3, 1, 1, 1, 0, 0, 3, 1, 0, 1, 0, 1,
            0, 0, 2, 2, 0, 0, 2, 0, 0, 2, 0, 0
        ], dtype=torch.long)

        g = Graph(row, col, edge_weight, num_node, node_type, edge_type, x=features, y=labels)
        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "official":
            labels = np.array([
                1, 1, 1, 1, 3, 3, 3, 1, 0, 1, 3, 1, 1, 1, 0, 0, 3, 1, 0, 1, 0, 1,
                0, 0, 2, 2, 0, 0, 2, 0, 0, 2, 0, 0
            ])

            # By Default: Select a single training node (the first one) for each community
            num_train_per_class = self._num_train_per_class
            num_val = self._num_valid_per_class

            train_idx, val_idx, test_idx = np.empty(0), np.empty(0), np.empty(0)
            for i in range(self.num_classes):
                idx = np.nonzero(labels == i)[0]
                train_idx = np.append(train_idx, idx[:num_train_per_class])
                val_idx = np.append(val_idx, idx[num_train_per_class: num_train_per_class + num_val])
                test_idx = np.append(test_idx, idx[num_train_per_class + num_val:])
            train_idx.reshape(-1)
            val_idx.reshape(-1)
            test_idx.reshape(-1)

        elif split == "random":
            raise NotImplementedError
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



from sgl.models.base_model import BaseSGAPModel
from sgl.models.simple_models import MultiLayerPerceptron
from sgl.operators.graph_op import LaplacianGraphOp
from sgl.operators.message_op import ConcatMessageOp, ProjectedConcatMessageOp


# slightly different from the original design
class SIGN(BaseSGAPModel):
    def __init__(self, prop_steps, feat_dim, output_dim, hidden_dim, num_layers):
        super(SIGN, self).__init__(prop_steps, feat_dim, output_dim)

        self._pre_graph_op = LaplacianGraphOp(prop_steps, r=0.5)
        # self._pre_msg_op = ProjectedConcatMessageOp(0, prop_steps + 1, feat_dim, hidden_dim)
        self._pre_msg_op = ConcatMessageOp(0, prop_steps + 1)
        self._base_model = MultiLayerPerceptron((prop_steps + 1) * feat_dim, hidden_dim, num_layers, output_dim)



from .node_classification import HeteroNodeClassification
from .node_classification import NodeClassification
from .node_clustering import NodeClustering
from .node_clustering import NodeClusteringNAFS
from .link_prediction import LinkPredictionGAE
from .link_prediction import LinkPredictionNAFS
from .correct_and_smooth import NodeClassification_With_CorrectAndSmooth
from .node_classification_with_label_use import NodeClassificationWithLabelUse

__all__ = [
    "NodeClassification",
    "HeteroNodeClassification",
    "NodeClustering",
    "NodeClusteringNAFS",
    "LinkPredictionGAE",
    "LinkPredictionNAFS",
    "NodeClassification_With_CorrectAndSmooth"
    "NodeClassificationWithLabelUse"
]



from sgl.models.base_model import BaseSGAPModel
from sgl.models.simple_models import ResMultiLayerPerceptron
from sgl.operators.graph_op import LaplacianGraphOp, PprGraphOp
from sgl.operators.message_op import LearnableWeightedMessageOp, LastMessageOp
        

class PASCA_V3(BaseSGAPModel):
    def __init__(self, prop_steps, post_steps, feat_dim, output_dim, hidden_dim, num_layers):
        super(PASCA_V3, self).__init__(prop_steps, feat_dim, output_dim)

        self._pre_graph_op = LaplacianGraphOp(prop_steps, r=0.5)
        self._pre_msg_op = LearnableWeightedMessageOp(1, prop_steps + 1, "gate", feat_dim)
        self._base_model = ResMultiLayerPerceptron(feat_dim, hidden_dim, num_layers, output_dim, 0.8)
        self._post_graph_op = PprGraphOp(post_steps, r=0.5, alpha=0.3)
        self._post_msg_op = LastMessageOp()



import os.path as osp
import numpy as np
import numpy.ctypeslib as ctl
import scipy.sparse as sp
import torch
from ctypes import c_int
from torch import Tensor


def csr_sparse_dense_matmul(adj, feature):
    file_path = osp.abspath(__file__)
    dir_path = osp.split(file_path)[0]

    ctl_lib = ctl.load_library("./csrc/libmatmul.so", dir_path)

    arr_1d_int = ctl.ndpointer(
        dtype=np.int32,
        ndim=1,
        flags="CONTIGUOUS"
    )

    arr_1d_float = ctl.ndpointer(
        dtype=np.float32,
        ndim=1,
        flags="CONTIGUOUS"
    )
    ctl_lib.FloatCSRMulDenseOMP.argtypes = [arr_1d_float, arr_1d_float, arr_1d_int, arr_1d_int, arr_1d_float,
                                            c_int, c_int]
    ctl_lib.FloatCSRMulDenseOMP.restypes = None

    answer = np.zeros(feature.shape).astype(np.float32).flatten()
    data = adj.data.astype(np.float32)
    indices = adj.indices
    indptr = adj.indptr
    mat = feature.flatten()
    mat_row, mat_col = feature.shape

    ctl_lib.FloatCSRMulDenseOMP(answer, data, indices, indptr, mat, mat_row, mat_col)

    return answer.reshape(feature.shape)


def cuda_csr_sparse_dense_matmul(adj, feature):
    file_path = osp.abspath(__file__)
    dir_path = osp.split(file_path)[0]
    
    ctl_lib = ctl.load_library("./csrc/libcudamatmul.so", dir_path)

    arr_1d_int = ctl.ndpointer(
        dtype=np.int32,
        ndim=1,
        flags="CONTIGUOUS"
    )
    arr_1d_float = ctl.ndpointer(
        dtype=np.float32,
        ndim=1,
        flags="CONTIGUOUS"
    )
    ctl_lib.FloatCSRMulDense.argtypes = [arr_1d_float, c_int, arr_1d_float, arr_1d_int, arr_1d_int, arr_1d_float, c_int,
                                         c_int]
    ctl_lib.FloatCSRMulDense.restypes = c_int

    answer = np.zeros(feature.shape).astype(np.float32).flatten()
    data = adj.data.astype(np.float32)
    data_nnz = len(data)
    indices = adj.indices
    indptr = adj.indptr
    mat = feature.flatten()
    mat_row, mat_col = feature.shape

    ctl_lib.FloatCSRMulDense(answer, data_nnz, data, indices, indptr, mat, mat_row, mat_col)

    return answer.reshape(feature.shape)


def adj_to_symmetric_norm(adj, r):
    adj = adj + sp.eye(adj.shape[0])
    degrees = np.array(adj.sum(1))
    r_inv_sqrt_left = np.power(degrees, r - 1).flatten()
    r_inv_sqrt_left[np.isinf(r_inv_sqrt_left)] = 0.
    r_mat_inv_sqrt_left = sp.diags(r_inv_sqrt_left)

    r_inv_sqrt_right = np.power(degrees, -r).flatten()
    r_inv_sqrt_right[np.isinf(r_inv_sqrt_right)] = 0.
    r_mat_inv_sqrt_right = sp.diags(r_inv_sqrt_right)

    adj_normalized = adj.dot(r_mat_inv_sqrt_left).transpose().dot(r_mat_inv_sqrt_right)
    return adj_normalized


def one_dim_weighted_add(feat_list, weight_list):
    if not isinstance(feat_list, list) or not isinstance(weight_list, Tensor):
        raise TypeError("This function is designed for list(feature) and tensor(weight)!")
    elif len(feat_list) != weight_list.shape[0]:
        raise ValueError("The feature list and the weight list have different lengths!")
    elif len(weight_list.shape) != 1:
        raise ValueError("The weight list should be a 1d tensor!")

    feat_shape = feat_list[0].shape
    feat_reshape = torch.vstack([feat.view(1, -1).squeeze(0) for feat in feat_list])
    weighted_feat = (feat_reshape * weight_list.view(-1, 1)).sum(dim=0).view(feat_shape)
    return weighted_feat


def two_dim_weighted_add(feat_list, weight_list):
    if not isinstance(feat_list, list) or not isinstance(weight_list, Tensor):
        raise TypeError("This function is designed for list(feature) and tensor(weight)!")
    elif len(feat_list) != weight_list.shape[1]:
        raise ValueError("The feature list and the weight list have different lengths!")
    elif len(weight_list.shape) != 2:
        raise ValueError("The weight list should be a 2d tensor!")

    feat_reshape = torch.stack(feat_list, dim=2)
    weight_reshape = weight_list.unsqueeze(dim=2)
    weighted_feat = torch.bmm(feat_reshape, weight_reshape).squeeze(dim=2)
    return weighted_feat



import heapq
import time
import torch
import warnings
from functools import reduce
from typing import Dict, List, Tuple

from sgl.dataset import Dblp
from sgl.models.hetero import NARS_SIGN, Fast_NARS_SGC_WithLearnableWeights
from sgl.tasks import HeteroNodeClassification
from sgl.utils import GpuWithMaxFreeMem

# Hyperparameters
PROP_STEPS = 3
HIDDEN_DIM = 256
NUM_LAYERS = 2
NUM_EPOCHS_TO_TRAIN = 50
NUM_EPOCHS_TO_FIND_WEIGHT = 20
LR = 0.01
WEIGHT_DECAY = 0.0
BATCH_SIZE = 10000


def GenerateSubgraphsWithSameEdgeTypeNum(dataset, random_subgraph_num: int, subgraph_edge_type_num: int) -> Dict:
    return dataset.nars_preprocess(edge_types=dataset.EDGE_TYPES,
                                   predict_class=dataset.TYPE_OF_NODE_TO_PREDICT,
                                   random_subgraph_num=random_subgraph_num,
                                   subgraph_edge_type_num=subgraph_edge_type_num)


# Input format: [(random_subgraph_num, subgraph_edge_type_num), ...]
# Each element is a tuple of (random_subgraph_num, subgraph_edge_type_num)
def GenerateSubgraphDict(dataset, subgraph_config: List) -> Dict:
    subgraph_list = [GenerateSubgraphsWithSameEdgeTypeNum(
        dataset, random_subgraph_num, subgraph_edge_type_num)
        for random_subgraph_num, subgraph_edge_type_num
        in subgraph_config]

    return reduce(lambda x, y: {**x, **y}, subgraph_list)


def Dict2List(dict: Dict) -> List:
    return [(key, dict[key]) for key in dict]


# Input format: [(random_subgraph_num, subgraph_edge_type_num), ...]
# Each element is a tuple of (random_subgraph_num, subgraph_edge_type_num)
def GenerateSubgraphList(dataset, subgraph_config: List) -> List:
    return Dict2List(GenerateSubgraphDict(dataset, subgraph_config))


# Input format: [(random_subgraph_num, subgraph_edge_type_num), ...]
# Each element is a tuple of (random_subgraph_num, subgraph_edge_type_num)
def OneTrialWithSubgraphConfig(dataset, subgraph_config: List, num_epochs: int) -> Tuple[
    float, List, torch.torch.Tensor]:
    subgraph_list = GenerateSubgraphList(dataset, subgraph_config)

    predict_class = dataset.TYPE_OF_NODE_TO_PREDICT

    model = Fast_NARS_SGC_WithLearnableWeights(prop_steps=PROP_STEPS,
                                               feat_dim=dataset.data.num_features[predict_class],
                                               output_dim=dataset.data.num_classes[predict_class],
                                               hidden_dim=HIDDEN_DIM, num_layers=NUM_LAYERS,
                                               random_subgraph_num=len(subgraph_list))

    device = torch.device(
        f"cuda:{GpuWithMaxFreeMem()}" if torch.cuda.is_available() else "cpu")
    classification = HeteroNodeClassification(dataset, predict_class, model,
                                              lr=LR, weight_decay=WEIGHT_DECAY,
                                              epochs=num_epochs, device=device,
                                              train_batch_size=BATCH_SIZE,
                                              eval_batch_size=BATCH_SIZE,
                                              subgraph_list=subgraph_list,
                                              seed=int(time.time()),
                                              record_subgraph_weight=True)
    test_acc = classification.test_acc
    raw_weight = classification.subgraph_weight
    weight_sum = raw_weight.sum()
    normalized_weight = raw_weight / weight_sum

    return test_acc, subgraph_list, normalized_weight


def TopKIndex(k: int, tensor: torch.Tensor) -> List:
    return heapq.nlargest(k, range(tensor.size(0)), key=lambda i: tensor[i])


def OneTrialWithSubgraphList(dataset, subgraph_list: List, num_epochs: int) -> Tuple[
    float, List, torch.Tensor]:
    predict_class = dataset.TYPE_OF_NODE_TO_PREDICT
    model = Fast_NARS_SGC_WithLearnableWeights(prop_steps=PROP_STEPS,
                                               feat_dim=dataset.data.num_features[predict_class],
                                               output_dim=dataset.data.num_classes[predict_class],
                                               hidden_dim=HIDDEN_DIM, num_layers=NUM_LAYERS,
                                               random_subgraph_num=len(subgraph_list))

    device = torch.device(
        f"cuda:{GpuWithMaxFreeMem()}" if torch.cuda.is_available() else "cpu")
    classification = HeteroNodeClassification(dataset, predict_class, model,
                                              lr=LR, weight_decay=WEIGHT_DECAY,
                                              epochs=num_epochs, device=device,
                                              train_batch_size=BATCH_SIZE,
                                              eval_batch_size=BATCH_SIZE,
                                              subgraph_list=subgraph_list,
                                              seed=int(time.time()),
                                              record_subgraph_weight=True)

    test_acc = classification.test_acc
    raw_weight = classification.subgraph_weight
    weight_sum = raw_weight.sum()
    normalized_weight = raw_weight / weight_sum

    return test_acc, subgraph_list, normalized_weight


# Input format: [(random_subgraph_num, subgraph_edge_type_num), ...]
# Each element is a tuple of (random_subgraph_num, subgraph_edge_type_num)
# Only top k subgraphs with highest weights are retained
def OneTrialWithSubgraphListTopK(dataset, subgraph_config: List, k: int,
                                 num_epochs_to_find_weight: int, num_epochs_to_train: int) -> float:
    original_test_acc, subgraph_list, normalized_weight = OneTrialWithSubgraphConfig(
        dataset, subgraph_config, num_epochs_to_find_weight)
    if k > len(subgraph_list):
        k = len(subgraph_list)
        warnings.warn('k is larger than the number of subgraphs,'
                      'k is set to the number of subgraphs',
                      UserWarning)

    top_k_index = TopKIndex(k, normalized_weight.abs())
    retained_subgraph_list = [subgraph_list[i] for i in top_k_index]

    test_acc, _, _ = OneTrialWithSubgraphList(
        dataset, retained_subgraph_list, num_epochs_to_train)

    return test_acc, original_test_acc


def main():
    dataset = Dblp(root='.', path_of_zip='./dataset/DBLP_processed.zip')

    SUBGRAPH_CONFIG = [(3, 2), (3, 3)]
    test_acc, original_test_acc = OneTrialWithSubgraphListTopK(dataset,
                                                               SUBGRAPH_CONFIG, 2,
                                                               num_epochs_to_train=NUM_EPOCHS_TO_TRAIN,
                                                               num_epochs_to_find_weight=NUM_EPOCHS_TO_FIND_WEIGHT)

    print('test_acc:', test_acc)
    print('original_test_acc:', original_test_acc)


if __name__ == '__main__':
    main()



from sgl.models.base_model import BaseSGAPModel
from sgl.models.simple_models import MultiLayerPerceptron
from sgl.operators.graph_op import LaplacianGraphOp
from sgl.operators.message_op import SimpleWeightedMessageOp


class GBP(BaseSGAPModel):
    def __init__(self, prop_steps, feat_dim, output_dim, hidden_dim, num_layers, r=0.5, alpha=0.85):
        super(GBP, self).__init__(prop_steps, feat_dim, output_dim)

        self._pre_graph_op = LaplacianGraphOp(prop_steps, r=0.5)
        self._pre_msg_op = SimpleWeightedMessageOp(0, prop_steps + 1, "alpha", alpha)
        self._base_model = MultiLayerPerceptron(feat_dim, hidden_dim, num_layers, output_dim)



from .gamlp import GAMLP
from .gamlp_recursive import GAMLPRecursive
from .gbp import GBP
from .sgc import SGC
from .sign import SIGN
from .ssgc import SSGC
from .nafs import NAFS

__all__ = [
    "SGC",
    "SIGN",
    "SSGC",
    "GBP",
    "GAMLP",
    "GAMLPRecursive",
    "NAFS"
]



import json
import numpy as np
import os.path as osp
import pickle as pkl
import torch
from itertools import chain

from sgl.data.base_data import Graph
from sgl.data.base_dataset import NodeDataset
from sgl.dataset.utils import pkl_read_file, download_to, to_undirected


class Wikics(NodeDataset):
    # Have 20 different split of training and validation set, identified by split_id in [0, 19]
    # Currently, we only support calculating the accuracy of one split, 
    # and average accuracy will be supported in the future.
    def __init__(self, name="wikics", root="./", split="official", is_undirected=True, split_id=0):
        if split_id not in range(20):
            raise ValueError("Split id not supported")

        self._is_undirected = is_undirected
        super(Wikics, self).__init__(root + "Wikics", name)

        self._split_id = split_id
        self._data = pkl_read_file(self.processed_file_paths)
        self._split = split
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(split)

    @property
    def raw_file_paths(self):
        filenames = ["data.json"]
        return [osp.join(self._raw_dir, filename) for filename in filenames]

    @property
    def processed_file_paths(self):
        name = self._name + "_undirected" if self._is_undirected else self._name
        return osp.join(self._processed_dir, f"{name}.graph")

    def _download(self):
        url = 'https://github.com/pmernyei/wiki-cs-dataset/raw/master/dataset'
        path = self.raw_file_paths[0]
        file_url = url + 'data.json'
        print(file_url)
        download_to(file_url, path)

    def _process(self):
        with open(self.raw_file_paths[0], 'r') as f:
            data = json.load(f)

        features = np.array(data['features'])
        labels = torch.LongTensor(data['labels'])
        num_node = features.shape[0]
        node_type = "article"

        edges = [[(i, j) for j in js] for i, js in enumerate(data['links'])]
        edges = list(chain(*edges))
        edge_index = torch.tensor(edges, dtype=torch.long).t().contiguous()
        if self._is_undirected:
            edge_index = to_undirected(edge_index)

        edge_index = edge_index.numpy()
        row, col = edge_index[0], edge_index[1]
        edge_weight = np.ones(len(row))
        edge_type = "article__to__article"

        g = Graph(row, col, edge_weight, num_node, node_type, edge_type, x=features, y=labels)
        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "official":
            with open(self.raw_file_paths[0], 'r') as f:
                data = json.load(f)

            train_mask = torch.tensor(data['train_masks'], dtype=torch.bool)
            train_mask = train_mask.contiguous()

            val_mask = torch.tensor(data['val_masks'], dtype=torch.bool)
            val_mask = val_mask.contiguous()

            stopping_mask = torch.tensor(data['stopping_masks'], dtype=torch.bool)
            stopping_mask = stopping_mask.contiguous()

            test_mask = torch.tensor(data['test_mask'], dtype=torch.bool)

            # Choose the specific split
            train_mask = train_mask[self._split_id]
            val_mask = val_mask[self._split_id]
            stopping_mask = stopping_mask[self._split_id]

            # Different from PyG, we merged the stopping_idx into valid_idx
            train_idx = torch.nonzero(train_mask == 1).reshape(-1)
            val_idx = torch.nonzero(val_mask == 1).reshape(-1)
            stopping_idx = torch.nonzero(stopping_mask == 1).reshape(-1)
            val_idx = torch.cat((val_idx, stopping_idx), dim=0)
            test_idx = torch.nonzero(test_mask == 1).reshape(-1)

        elif split == "random":
            raise NotImplementedError
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



import time
import torch
import torch.nn as nn
from torch.optim import Adam
from torch.utils.data import DataLoader
import torch.nn.functional as F

from sgl.tasks.base_task import BaseTask
from sgl.tasks.utils import accuracy, set_seed, train, mini_batch_train, evaluate, \
                            mini_batch_evaluate, adj_to_symmetric_norm
from sgl.tricks import CorrectAndSmooth

class NodeClassification_With_CorrectAndSmooth(BaseTask):
    def __init__(self, dataset, model, lr, weight_decay, epochs, device, num_correct_layers, 
                 correct_alpha, num_smooth_layers, smooth_alpha, autoscale=True, scale=1.0, 
                 loss_fn=nn.CrossEntropyLoss(), seed=42, train_batch_size=None, eval_batch_size=None, 
                 correct_r=0.5, smooth_r=0.5):
        super(NodeClassification_With_CorrectAndSmooth, self).__init__()

        self.__dataset = dataset
        self.__labels = self.__dataset.y

        self.__model = model
        self.__optimizer = Adam(model.parameters(), lr=lr,
                                weight_decay=weight_decay)
        self.__epochs = epochs
        self.__loss_fn = loss_fn
        self.__device = device
        self.__seed = seed

        self.__post_trick = CorrectAndSmooth(num_correct_layers, correct_alpha, 
                                        num_smooth_layers, smooth_alpha, autoscale, scale)
        self.__smooth_r = smooth_r
        self.__correct_r = correct_r

        self.__mini_batch = False
        if train_batch_size is not None:
            self.__mini_batch = True
            self.__train_loader = DataLoader(
                self.__dataset.train_idx, batch_size=train_batch_size, shuffle=True, drop_last=False)
            self.__val_loader = DataLoader(
                self.__dataset.val_idx, batch_size=eval_batch_size, shuffle=False, drop_last=False)
            self.__test_loader = DataLoader(
                self.__dataset.test_idx, batch_size=eval_batch_size, shuffle=False, drop_last=False)
            self.__all_eval_loader = DataLoader(
                range(self.__dataset.data.num_node), batch_size=eval_batch_size, shuffle=False, drop_last=False)

        self.__test_acc = self._execute()

    @property
    def test_acc(self):
        return self.__test_acc

    def _execute(self):
        set_seed(self.__seed)

        pre_time_st = time.time()
        self.__model.preprocess(self.__dataset.adj, self.__dataset.x)
        pre_time_ed = time.time()
        print(f"Preprocessing done in {(pre_time_ed - pre_time_st):.4f}s")

        self.__model = self.__model.to(self.__device)
        self.__labels = self.__labels.to(self.__device)

        t_total = time.time()
        best_val = 0.
        best_test = 0.
        best_y_soft = None
        for epoch in range(self.__epochs):
            t = time.time()
            if self.__mini_batch is False:
                loss_train, acc_train = train(self.__model, self.__dataset.train_idx, self.__labels, self.__device,
                                              self.__optimizer, self.__loss_fn)
                acc_val, acc_test = evaluate(self.__model, self.__dataset.val_idx, self.__dataset.test_idx,
                                             self.__labels, self.__device)
            else:
                loss_train, acc_train = mini_batch_train(self.__model, self.__dataset.train_idx, self.__train_loader,
                                                         self.__labels, self.__device, self.__optimizer, self.__loss_fn)
                acc_val, acc_test = mini_batch_evaluate(self.__model, self.__dataset.val_idx, self.__val_loader,
                                                        self.__dataset.test_idx, self.__test_loader, self.__labels,
                                                        self.__device)

            print('Epoch: {:03d}'.format(epoch + 1),
                  'loss_train: {:.4f}'.format(loss_train),
                  'acc_train: {:.4f}'.format(acc_train),
                  'acc_val: {:.4f}'.format(acc_val),
                  'acc_test: {:.4f}'.format(acc_test),
                  'time: {:.4f}s'.format(time.time() - t))
            if acc_val > best_val:
                idx = torch.arange(0, self.__dataset.num_node)
                output = self.__model.model_forward(idx, self.__device).detach()
                best_y_soft = F.softmax(output, dim=1)
                best_val = acc_val
                best_test = acc_test

        acc_val, acc_test = self._postprocess(best_y_soft)
        
        print(f"After C&S, acc_val: {acc_val:.4f} acc_test: {acc_test:.4f}")
        if acc_val > best_val:
            best_val = acc_val
            best_test = acc_test

        print("Optimization Finished!")
        print("Total time elapsed: {:.4f}s".format(time.time() - t_total))
        print(f'Best val: {best_val:.4f}, best test: {best_test:.4f}')
        return best_test

    def _postprocess(self, y_soft):
        self.__model.eval()

        correct_adj = adj_to_symmetric_norm(adj=self.__dataset.adj, r=self.__correct_r)
        smooth_adj = adj_to_symmetric_norm(adj=self.__dataset.adj, r=self.__smooth_r)
        
        post = self.__post_trick
        final_output = post.correct(y_soft, self.__labels, self.__dataset.train_idx, correct_adj)
        final_output = post.smooth(final_output, self.__labels, self.__dataset.train_idx, smooth_adj)

        acc_val = accuracy(
            final_output[self.__dataset.val_idx], self.__labels[self.__dataset.val_idx])
        acc_test = accuracy(
            final_output[self.__dataset.test_idx], self.__labels[self.__dataset.test_idx])
        return acc_val, acc_test



import numpy as np
import os
import os.path as osp
import pickle as pkl
import scipy.sparse as sp
import torch
from torch_geometric.data import extract_zip

from sgl.data.base_data import Graph
from sgl.data.base_dataset import NodeDataset
from sgl.dataset.utils import pkl_read_file, download_to


class Reddit(NodeDataset):
    def __init__(self, name="reddit", root="./", split="official"):
        if name not in ["reddit"]:
            raise ValueError("Dataset name not supported!")
        super(Reddit, self).__init__(root + "Reddit/", name)

        self._data = pkl_read_file(self.processed_file_paths)
        self._split = split
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(
            split)

    @property
    def raw_file_paths(self):
        filenames = ["reddit_data.npz", "reddit_graph.npz"]
        return [osp.join(self._raw_dir, filename) for filename in filenames]

    @property
    def processed_file_paths(self):
        filename = "graph"
        return osp.join(self._processed_dir, "{}.{}".format(self._name, filename))

    def _download(self):
        url = "https://data.dgl.ai/dataset/reddit.zip"
        path = osp.join(self._raw_dir, "reddit.zip")
        print(url)
        download_to(url, path)
        extract_zip(path, self._raw_dir)
        os.unlink(path)

    def _process(self):
        adj = sp.coo_matrix(sp.load_npz(
            osp.join(self._raw_dir, 'reddit_graph.npz')))
        row, col, edge_weight = adj.row, adj.col, adj.data
        edge_type = "post__to__post"

        data = np.load(osp.join(self._raw_dir, 'reddit_data.npz'))
        features, labels, split = data['feature'], data['label'], data['node_types']
        num_node = features.shape[0]
        node_type = "post"
        labels = torch.LongTensor(labels)

        g = Graph(row, col, edge_weight, num_node,
                  node_type, edge_type, x=features, y=labels)
        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "official":
            data = np.load(osp.join(self._raw_dir, 'reddit_data.npz'))
            split = torch.tensor(data['node_types'])

            train_idx = torch.nonzero(split == 1).reshape(-1)
            val_idx = torch.nonzero(split == 2).reshape(-1)
            test_idx = torch.nonzero(split == 3).reshape(-1)
        elif split == "random":
            raise NotImplementedError
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



import torch
import torch.nn.functional as F
from sgl.tricks.utils import label_propagation


class CorrectAndSmooth:
    def __init__(self, num_correct_layers, correct_alpha, num_smooth_layers, 
                smooth_alpha, autoscale=True, scale=1.0) -> None:
        super().__init__()
        self.__num_correct_layers = num_correct_layers
        self.__correct_alpha = correct_alpha
        self.__num_smooth_layers = num_smooth_layers
        self.__smooth_alpha = smooth_alpha
        self.__autoscale = autoscale
        self.__scale = scale

    # different from pyg implemetation, y_true here represents all the labels for convenience
    @torch.no_grad()
    def correct(self, y_soft, y_true, mask, adj):
        y_soft = y_soft.cpu()
        y_true = y_true.cpu()
        mask = torch.tensor(mask)
        if y_true.dtype == torch.long:
            y_true = F.one_hot(y_true.view(-1), y_soft.size(-1))
            y_true = y_true.to(y_soft.dtype)

        error = torch.zeros_like(y_soft)
        error[mask] = y_true[mask] - y_soft[mask]
        num_true = mask.shape[0] if mask.dtype == torch.long else int(mask.sum())

        if self.__autoscale:
            smoothed_error = label_propagation(error, adj, self.__num_correct_layers, self.__correct_alpha, post_process=lambda x:x.clamp_(-1., 1.))
            sigma = error[mask].abs().sum() / num_true
            scale = sigma / smoothed_error.abs().sum(dim=1, keepdim=True)
            scale[scale.isinf() | (scale > 1000)] = 1.0
            return y_soft + smoothed_error * scale
        
        else:
            def fix_input(x):
                x[mask] = error[mask]
                return x

            smoothed_error = label_propagation(error, adj, self.__num_correct_layers, self.__correct_alpha, post_process=fix_input)
            return y_soft + smoothed_error * self.__scale 

    @torch.no_grad()
    def smooth(self, y_soft, y_true, mask, adj):
        y_soft = y_soft.cpu()
        y_true = y_true.cpu()
        if y_true.dtype == torch.long:
            y_true = F.one_hot(y_true.view(-1), y_soft.size(-1))
            y_true = y_true.to(y_soft.dtype)
        
        y_soft[mask] = y_true[mask]

        smoothed_label = label_propagation(y_soft, adj, self.__num_smooth_layers, self.__smooth_alpha)
        return smoothed_label



import torch
import torch.nn as nn
import torch.nn.functional as F

from sgl.data.base_dataset import HeteroNodeDataset


class BaseSGAPModel(nn.Module):
    def __init__(self, prop_steps, feat_dim, output_dim):
        super(BaseSGAPModel, self).__init__()
        self._prop_steps = prop_steps
        self._feat_dim = feat_dim
        self._output_dim = output_dim

        self._pre_graph_op, self._pre_msg_op = None, None
        self._post_graph_op, self._post_msg_op = None, None
        self._base_model = None

        self._processed_feat_list = None
        self._processed_feature = None
        self._pre_msg_learnable = False

    def preprocess(self, adj, feature):
        if self._pre_graph_op is not None:
            self._processed_feat_list = self._pre_graph_op.propagate(
                adj, feature)
            if self._pre_msg_op.aggr_type in [
                "proj_concat", "learnable_weighted", "iterate_learnable_weighted"]:
                self._pre_msg_learnable = True
            else:
                self._pre_msg_learnable = False
                self._processed_feature = self._pre_msg_op.aggregate(
                    self._processed_feat_list)
        else:
            self._pre_msg_learnable = False
            self._processed_feature = feature

    def postprocess(self, adj, output):
        if self._post_graph_op is not None:
            if self._post_msg_op.aggr_type in [
                "proj_concat", "learnable_weighted", "iterate_learnable_weighted"]:
                raise ValueError(
                    "Learnable weighted message operator is not supported in the post-processing phase!")
            output = F.softmax(output, dim=1)
            output = output.detach().numpy()
            output = self._post_graph_op.propagate(adj, output)
            output = self._post_msg_op.aggregate(output)

        return output

    # a wrapper of the forward function
    def model_forward(self, idx, device):
        return self.forward(idx, device)

    def forward(self, idx, device):
        processed_feature = None
        if self._pre_msg_learnable is False:
            processed_feature = self._processed_feature[idx].to(device)
        else:
            transferred_feat_list = [feat[idx].to(
                device) for feat in self._processed_feat_list]
            processed_feature = self._pre_msg_op.aggregate(
                transferred_feat_list)

        output = self._base_model(processed_feature)
        return output


class BaseHeteroSGAPModel(nn.Module):
    def __init__(self, prop_steps, feat_dim, output_dim):
        super(BaseHeteroSGAPModel, self).__init__()
        self._prop_steps = prop_steps
        self._feat_dim = feat_dim
        self._output_dim = output_dim

        self._pre_graph_op, self._pre_msg_op = None, None
        self._aggregator = None
        self._base_model = None

        self._propagated_feat_list_list = None
        self._processed_feature_list = None
        self._pre_msg_learnable = False

    # Either subgraph_list or (random_subgraph_num, subgraph_edge_type_num) should be provided.
    def preprocess(self, dataset, predict_class,
                   random_subgraph_num=-1, subgraph_edge_type_num=-1,
                   subgraph_list=None):
        if subgraph_list is None and (random_subgraph_num == -1 or subgraph_edge_type_num == -1):
            raise ValueError(
                "Either subgraph_list or (random_subgraph_num, subgraph_edge_type_num) should be provided!")
        if subgraph_list is not None and (random_subgraph_num != -1 or subgraph_edge_type_num != -1):
            raise ValueError(
                "subgraph_list is provided, random_subgraph_num and subgraph_edge_type_num will be ignored!")

        if not isinstance(dataset, HeteroNodeDataset):
            raise TypeError(
                "Dataset must be an instance of HeteroNodeDataset!")
        elif predict_class not in dataset.node_types:
            raise ValueError("Please input valid node class for prediction!")
        predict_idx = dataset.data.node_id_dict[predict_class]

        if subgraph_list is None:
            subgraph_dict = dataset.nars_preprocess(dataset.edge_types, predict_class,
                                                    random_subgraph_num,
                                                    subgraph_edge_type_num)
            subgraph_list = [(key, subgraph_dict[key])
                             for key in subgraph_dict]

        self._propagated_feat_list_list = [[]
                                           for _ in range(self._prop_steps + 1)]

        for key, value in subgraph_list:
            edge_type_list = []
            for edge_type in key:
                edge_type_list.append(edge_type.split("__")[0])
                edge_type_list.append(edge_type.split("__")[2])
            if predict_class in edge_type_list:
                adj, feature, node_id = value
                propagated_feature = self._pre_graph_op.propagate(adj, feature)

                start_pos = list(node_id).index(predict_idx[0])
                for i, feature in enumerate(propagated_feature):
                    self._propagated_feat_list_list[i].append(
                        feature[start_pos:start_pos + dataset.data.num_node[predict_class]])

    # a wrapper of the forward function
    def model_forward(self, idx, device):
        return self.forward(idx, device)

    def forward(self, idx, device):
        feat_input = []
        for x_list in self._propagated_feat_list_list:
            feat_input.append([])
            for x in x_list:
                feat_input[-1].append(x[idx].to(device))

        aggregated_feat_list = self._aggregator(feat_input)
        combined_feat = self._pre_msg_op.aggregate(aggregated_feat_list)
        output = self._base_model(combined_feat)

        return output


class FastBaseHeteroSGAPModel(nn.Module):
    def __init__(self, prop_steps, feat_dim, output_dim):
        super(FastBaseHeteroSGAPModel, self).__init__()
        self._prop_steps = prop_steps
        self._feat_dim = feat_dim
        self._output_dim = output_dim

        self._pre_graph_op = None
        self._aggregator = None
        self._base_model = None

        self._propagated_feat_list_list = None
        self._processed_feature_list = None
        self._pre_msg_learnable = False

    # Either subgraph_list or (random_subgraph_num, subgraph_edge_type_num) should be provided.
    def preprocess(self, dataset, predict_class,
                   random_subgraph_num=-1, subgraph_edge_type_num=-1,
                   subgraph_list=None):
        if subgraph_list is None and (random_subgraph_num == -1 or subgraph_edge_type_num == -1):
            raise ValueError(
                "Either subgraph_list or (random_subgraph_num, subgraph_edge_type_num) should be provided!")
        if subgraph_list is not None and (random_subgraph_num != -1 or subgraph_edge_type_num != -1):
            raise ValueError(
                "subgraph_list is provided, random_subgraph_num and subgraph_edge_type_num will be ignored!")

        if not isinstance(dataset, HeteroNodeDataset):
            raise TypeError(
                "Dataset must be an instance of HeteroNodeDataset!")
        elif predict_class not in dataset.node_types:
            raise ValueError("Please input valid node class for prediction!")
        predict_idx = dataset.data.node_id_dict[predict_class]

        if subgraph_list is None:
            subgraph_dict = dataset.nars_preprocess(dataset.edge_types, predict_class,
                                                    random_subgraph_num,
                                                    subgraph_edge_type_num)
            subgraph_list = [(key, subgraph_dict[key])
                             for key in subgraph_dict]

        self._propagated_feat_list_list = [[]
                                           for _ in range(self._prop_steps + 1)]

        for key, value in subgraph_list:
            edge_type_list = []
            for edge_type in key:
                edge_type_list.append(edge_type.split("__")[0])
                edge_type_list.append(edge_type.split("__")[2])
            if predict_class in edge_type_list:
                adj, feature, node_id = value
                propagated_feature = self._pre_graph_op.propagate(adj, feature)

                start_pos = list(node_id).index(predict_idx[0])
                for i, feature in enumerate(propagated_feature):
                    self._propagated_feat_list_list[i].append(
                        feature[start_pos:start_pos + dataset.data.num_node[predict_class]])

        # 2-d list to 4-d tensor (num_node, feat_dim, num_subgraphs, prop_steps)
        self._propagated_feat_list_list = [torch.stack(
            x, dim=2) for x in self._propagated_feat_list_list]
        self._propagated_feat_list_list = torch.stack(
            self._propagated_feat_list_list, dim=3)

        # 4-d tensor to 3-d tensor (num_node, feat_dim, num_subgraphs * prop_steps)
        shape = self._propagated_feat_list_list.size()
        self._propagated_feat_list_list = self._propagated_feat_list_list.view(
            shape[0], shape[1], shape[2] * shape[3])

    # a wrapper of the forward function
    def model_forward(self, idx, device):
        return self.forward(idx, device)

    def forward(self, idx, device):
        feat_input = self._propagated_feat_list_list[idx].to(device)

        aggregated_feat_from_diff_hops = self._aggregator(feat_input)
        output = self._base_model(aggregated_feat_from_diff_hops)

        return output



from .concat_message_op import ConcatMessageOp
from .iterate_learnable_weighted_message_op import IterateLearnableWeightedMessageOp
from .last_message_op import LastMessageOp
from .learnable_weighted_messahe_op import LearnableWeightedMessageOp
from .max_message_op import MaxMessageOp
from .mean_message_op import MeanMessageOp
from .min_message_op import MinMessageOp
from .projected_concat_message_op import ProjectedConcatMessageOp
from .simple_weighted_message_op import SimpleWeightedMessageOp
from .sum_message_op import SumMessageOp
from .over_smooth_distance_op import OverSmoothDistanceWeightedOp

__all__ = [
    "ConcatMessageOp",
    "IterateLearnableWeightedMessageOp",
    "LastMessageOp",
    "LearnableWeightedMessageOp",
    "MaxMessageOp",
    "MeanMessageOp",
    "MinMessageOp",
    "ProjectedConcatMessageOp",
    "SimpleWeightedMessageOp",
    "SumMessageOp",
    "OverSmoothDistanceWeightedOp"
]



import scipy.sparse as sp

from sgl.operators.base_op import GraphOp
from sgl.operators.utils import adj_to_symmetric_norm


class PprGraphOp(GraphOp):
    def __init__(self, prop_steps, r=0.5, alpha=0.15):
        super(PprGraphOp, self).__init__(prop_steps)
        self.__r = r
        self.__alpha = alpha

    def _construct_adj(self, adj):
        if isinstance(adj, sp.csr_matrix):
            adj = adj.tocoo()
        elif not isinstance(adj, sp.coo_matrix):
            raise TypeError("The adjacency matrix must be a scipy.sparse.coo_matrix/csr_matrix!")

        adj_normalized = adj_to_symmetric_norm(adj, self.__r)
        adj_normalized = (1 - self.__alpha) * adj_normalized + self.__alpha * sp.eye(adj.shape[0])
        return adj_normalized.tocsr()



import time
import torch
from torch.optim import Adam
import torch.nn.functional as F
from sklearn.cluster import KMeans

from sgl.tasks.base_task import BaseTask
from sgl.tasks.utils import set_seed, clustering_train, cluster_loss, \
    sparse_mx_to_torch_sparse_tensor, adj_to_symmetric_norm
from sgl.tasks.clustering_metrics import clustering_metrics

class NodeClustering(BaseTask):
    def __init__(self, dataset, model, lr, weight_decay, epochs, device, loss_fn=cluster_loss, seed=42,
                 train_batch_size=None, eval_batch_size=None, n_init=20):
        super(NodeClustering, self).__init__()

        # clustering task does not support batch training
        if train_batch_size is not None or eval_batch_size is not None:
            raise ValueError("clustering task does not support batch training")
        
        self.__dataset = dataset
        self.__labels = self.__dataset.y

        self.__model = model
        self.__optimizer = Adam(model.parameters(), lr=lr,
                                weight_decay=weight_decay)
        self.__epochs = epochs
        self.__loss_fn = loss_fn
        self.__device = device
        self.__seed = seed

        # clustering task does not need valid set
        self.__cluster_train_idx = torch.arange(0, self.__dataset.num_node, dtype=torch.long)
        
        # params for Kmeans
        # note that the n_clusters should be equal to the number of different labels
        self.__n_clusters = self.__dataset.num_classes
        self.__n_init = n_init

        self.__acc, self.__nmi, self.__adjscore = self._execute()

    @property
    def acc(self):
        return self.__acc
    
    @property
    def nmi(self):
        return self.__nmi
    
    @property
    def adjscore(self):
        return self.__adjscore

    def _execute(self):
        set_seed(self.__seed)

        pre_time_st = time.time()
        self.__model.preprocess(self.__dataset.adj, self.__dataset.x)
        pre_time_ed = time.time()
        print(f"Preprocessing done in {(pre_time_ed - pre_time_st):.4f}s")

        self.__model = self.__model.to(self.__device)
        self.__labels = self.__labels.to(self.__device)

        t_total = time.time()
        best_acc = 0.
        best_nmi = 0.
        best_adjscore = 0.

        for epoch in range(self.__epochs):
            t = time.time()

            loss_train, acc, nmi, adjscore = clustering_train(self.__model, self.__cluster_train_idx, self.__labels,
            self.__device, self.__optimizer, self.__loss_fn, self.__n_clusters, self.__n_init)
                
            print("Epoch: {:03d}".format(epoch + 1),
                  "loss_train: {:.4f}".format(loss_train),
                  "acc: {:.4f}".format(acc),
                  "nmi: {:.4f}".format(nmi),
                  "adjscore: {:.4f}".format(adjscore),
                  "time: {:.4f}s".format(time.time() - t)
                )
            
            if acc > best_acc:
                best_acc = acc
            if nmi > best_nmi:
                best_nmi = nmi
            if adjscore > best_adjscore:
                best_adjscore = adjscore

        # postprocess
        acc, nmi, adjscore = self._postprocess()
        if acc > best_acc:
            best_acc = acc
        if nmi > best_nmi:
            best_nmi = nmi
        if adjscore > best_adjscore:
            best_adjscore = adjscore

        print("Optimization Finished!")
        print("Total time elapsed: {:.4f}s".format(time.time() - t_total))
        print(f"Best acc: {best_acc:.4f}, best_nmi: {best_nmi:.4f}, best_adjscore: {best_adjscore:.4f}")
        
        return best_acc, best_nmi, best_adjscore

    def _postprocess(self):
        self.__model.eval()
        outputs = self.__model.model_forward(
            range(self.__dataset.num_node), self.__device).to("cpu")
        
        final_output = self.__model.postprocess(outputs)
        kmeans = KMeans(n_clusters=self.__n_clusters, n_init=self.__n_init)
        y_pred = kmeans.fit_predict(final_output.data.cpu().numpy()) # cluster_label
        
        labels = self.__labels.cpu().numpy()
        cm = clustering_metrics(labels, y_pred)
        acc, nmi, adjscore = cm.evaluationClusterModelFromLabel()

        return acc, nmi, adjscore
        
class NodeClusteringNAFS(BaseTask):
    """ NodeClustering Tasks based on NAFS
        Fast and Does Not need train
    """
    def __init__(self, dataset, hops=20, method='mean', seed=42, n_init=20, r_list=[0.5, 0.4, 0.3, 0.2, 0.1, 0]):
        # note that: hops can be either an integer or a list
        # when it's a list, it contains the numbers of hops that you want to test
        # when it's a numebr, it will test all the hops in range(hops) 
        method = method.lower()
        if method not in ['mean', 'max', 'concat', 'simple']:
            raise ValueError("Method not Suppoted! Choose 'mean', 'max' or 'concat' !")    
        
        super(NodeClusteringNAFS, self).__init__()
        self.__dataset = dataset
        self.__labels = self.__dataset.y

        self.__method = method
        self.__r_list = r_list
        self.__hops = range(hops) if type(hops) == int else hops
        self.__seed = seed

        self.__n_clusters = self.__dataset.num_classes
        self.__n_init = n_init

        self.__best_hop_acc, self.__best_hop_nmi, self.__best_hop_adjscore, self.__acc, self.__nmi, self.__adjscore = self._execute()

    @property
    def acc(self):
        return self.__acc
    
    @property
    def nmi(self):
        return self.__nmi
    
    @property
    def adjscore(self):
        return self.__adjscore

    @property
    def best_hop_acc(self):
        return self.__best_hop_acc

    @property
    def best_hop_nmi(self):
        return self.__best_hop_nmi

    @property
    def best_hop_adjscore(self):
        return self.__best_hop_adjscore

    def _execute(self):
        set_seed(self.__seed)

        t_total = time.time()
        best_acc, best_nmi, best_adjscore = 0., 0. ,0.
        best_hop_acc, best_hop_nmi, best_hop_adjscore = 0, 0, 0

        for hop in self.__hops:
            t = time.time()
            acc, nmi, adjscore = self._k_hop_cluster(hop)

            print('hops:{:2d}'.format(hop),
                  'acc: {:.4f}'.format(acc),
                  'nmi: {:.4f}'.format(nmi),
                  'adjscore: {:.4f}'.format(adjscore),
                  'time: {:.4f} seconds'.format(time.time() - t)
                  )

            if acc > best_acc:
                best_acc = acc
                best_hop_acc = hop
            if nmi > best_nmi:
                best_nmi = nmi
                best_hop_nmi = hop
            if adjscore > best_adjscore:
                best_adjscore = adjscore
                best_hop_adjscore = hop

        print("Node Smoothing Finished!")
        print("Total time elapsed: {:.4f}s".format(time.time() - t_total))
        print("best_acc: {:.4f}, best_nmi: {:.4f}, best_adjscore: {:.4f}".format(best_acc, best_nmi, best_adjscore))

        return  best_hop_acc, best_hop_nmi, best_hop_adjscore, best_acc, best_nmi, best_adjscore

    def _k_hop_cluster(self, hops):
        input_features = []
        features = torch.tensor(self.__dataset.x, dtype=torch.float)
        r_list = self.__r_list
        for r in r_list:
            t = time.time()
            adj_norm = adj_to_symmetric_norm(self.__dataset.adj, r)
            adj_norm = sparse_mx_to_torch_sparse_tensor(adj_norm)
            features_list = []
            features_list.append(features)
            for _ in range(hops):
                features_list.append(torch.spmm(adj_norm, features_list[-1]))

            if self.__method == 'simple':
                input_features.append(features_list[-1])
                break

            weight_list = []
            norm_fea = torch.norm(features, 2, 1).add(1e-10)
            for fea in features_list:
                norm_cur = torch.norm(fea, 2, 1).add(1e-10)
                tmp = torch.div((features * fea).sum(1), norm_cur)
                tmp = torch.div(tmp, norm_fea)

                weight_list.append(tmp.unsqueeze(-1))

            tmp_weight = torch.cat(weight_list, dim=1)
            weight = F.softmax(tmp_weight, dim=1)

            input_feas = []
            for i in range(self.__dataset.num_node):
                fea = 0.
                for j in range(hops + 1):
                    fea += (weight[i][j]*features_list[j][i]).unsqueeze(0)
                input_feas.append(fea)
            input_feas = torch.cat(input_feas, dim=0)
            input_features.append(input_feas)
            print("hops {:d}: r={:.2f} calculation finished in {:.4f} seconds".format(hops, r, time.time() - t))
        
        if self.__method == 'mean':
            input_features = sum(input_features) / len(input_features)
        elif self.__method == 'max':
            input_features =  list(map(lambda x: x.unsqueeze(0), input_features))
            input_features = torch.cat(input_features, dim=0).max(0)[0]
        elif self.__method == 'concat':
            input_features = torch.cat(input_features, dim=1)
        elif self.__method == 'simple':
            input_features = input_features[-1]

        kmeans = KMeans(n_clusters=self.__n_clusters, n_init=self.__n_init, random_state=self.__seed)
        y_pred = kmeans.fit_predict(input_features.numpy())
        cm = clustering_metrics(self.__labels.numpy(), y_pred)
        acc, nmi, adjscore = cm.evaluationClusterModelFromLabel()
        return acc, nmi, adjscore



from sgl.models.base_model import BaseSGAPModel
from sgl.models.simple_models import LogisticRegression
from sgl.operators.graph_op import LaplacianGraphOp
from sgl.operators.message_op import MeanMessageOp


class SSGC(BaseSGAPModel):
    def __init__(self, prop_steps, feat_dim, output_dim):
        super(SSGC, self).__init__(prop_steps, feat_dim, output_dim)

        self._pre_graph_op = LaplacianGraphOp(prop_steps, r=0.5)
        self._pre_msg_op = MeanMessageOp(start=0, end=prop_steps + 1)
        self._base_model = LogisticRegression(feat_dim, output_dim)



import numpy as np
import os.path as osp
import pickle as pkl
import scipy.sparse as sp
import ssl
import sys
import torch
import urllib


def to_undirected(edge_index):
    row, col = edge_index
    new_row = torch.hstack((row, col))
    new_col = torch.hstack((col, row))
    new_edge_index = torch.stack((new_row, new_col), dim=0)

    return new_edge_index


def remove_self_loops(edge_index):
    mask = (edge_index[0] != edge_index[1])
    new_edge_index = edge_index[:, mask]
    return new_edge_index


def download_to(url, path):
    context = ssl._create_unverified_context()
    data = urllib.request.urlopen(url, context=context)

    with open(path, 'wb') as wf:
        try:
            wf.write(data.read())
        except IOError as e:
            print(e)
            exit(1)


def file_exist(filepaths):
    if isinstance(filepaths, list):
        for filepath in filepaths:
            if not osp.exists(filepath):
                return False
        return True
    else:
        if osp.exists(filepaths):
            return True
        else:
            return False


def pkl_read_file(filepath):
    file = None
    with open(filepath, 'rb') as rf:
        try:
            if sys.version_info > (3, 0):
                file = pkl.load(rf, encoding="latin1")
            else:
                file = pkl.load(rf)
        except IOError as e:
            print(e)
            exit(1)
    return file


def read_npz(path):
    with np.load(path) as f:
        return parse_npz(f)


def parse_npz(f):
    x = sp.csr_matrix((f['attr_data'], f['attr_indices'], f['attr_indptr']), f['attr_shape']).todense()
    x = torch.from_numpy(x).to(torch.float)
    x[x > 0] = 1

    adj = sp.csr_matrix((f['adj_data'], f['adj_indices'], f['adj_indptr']), f['adj_shape']).tocoo()
    row = torch.from_numpy(adj.row).to(torch.long)
    col = torch.from_numpy(adj.col).to(torch.long)
    edge_index = torch.stack([row, col], dim=0)
    edge_index = remove_self_loops(edge_index)
    edge_index = to_undirected(edge_index)

    y = torch.from_numpy(f['labels']).to(torch.long)

    return x, edge_index, y


def random_split_dataset(n_samples):
    val_idx = np.random.choice(list(range(n_samples)), size=int(n_samples * 0.2), replace=False)
    remain = set(range(n_samples)) - set(val_idx)
    test_idx = np.random.choice(list(remain), size=int(n_samples * 0.2), replace=False)
    train_idx = list(remain - set(test_idx))

    return train_idx, val_idx, test_idx



import numpy as np
import os.path as osp
import pickle as pkl
import torch

from sgl.data.base_data import Graph
from sgl.data.base_dataset import NodeDataset
from sgl.dataset.utils import download_to, pkl_read_file, random_split_dataset


class Github(NodeDataset):
    def __init__(self, name="github", root="./", split="official", num_train_per_class=30, num_valid_per_class=100):
        super(Github, self).__init__(root + "Github/", name)

        self._data = pkl_read_file(self.processed_file_paths)
        self._split = split
        self._num_train_per_class = num_train_per_class
        self._num_valid_per_class = num_valid_per_class
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(split)

    @property
    def raw_file_paths(self):
        filenames = ["github.npz"]
        return [osp.join(self._raw_dir, filename) for filename in filenames]

    @property
    def processed_file_paths(self):
        filename = "graph"
        return osp.join(self._processed_dir, "{}.{}".format(self._name, filename))

    def _download(self):
        url = "https://graphmining.ai/datasets/ptg/github.npz"
        path = osp.join(self._raw_dir, "github.npz")
        print(url)
        download_to(url, path)

    def _process(self):
        data = np.load(self.raw_file_paths[0])
        features = data["features"]
        num_node = features.shape[0]
        node_type = "developer"

        labels = data["target"]
        labels = torch.LongTensor(labels)

        edge_index = data["edges"].T
        row, col = edge_index[0], edge_index[1]
        edge_type = "developer__to__developer"

        # Default Edge weights
        edge_weight = np.ones(len(row))

        g = Graph(row, col, edge_weight, num_node, node_type, edge_type, x=features, y=labels)
        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "official":
            data = np.load(self.raw_file_paths[0])
            labels = data["target"]

            num_train_per_class = self._num_train_per_class
            num_val = self._num_valid_per_class
            train_idx, val_idx, test_idx = np.empty(0), np.empty(0), np.empty(0)
            for i in range(self.num_classes):
                idx = np.nonzero(labels == i)[0]
                train_idx = np.append(train_idx, idx[:num_train_per_class])
                val_idx = np.append(val_idx, idx[num_train_per_class: num_train_per_class + num_val])
                test_idx = np.append(test_idx, idx[num_train_per_class + num_val:])
            train_idx.reshape(-1)
            val_idx.reshape(-1)
            test_idx.reshape(-1)

        elif split == "random":
            train_idx, val_idx, test_idx = random_split_dataset(self.num_node)
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



import scipy.sparse as sp

from sgl.operators.base_op import GraphOp
from sgl.operators.utils import adj_to_symmetric_norm


class LaplacianGraphOp(GraphOp):
    def __init__(self, prop_steps, r=0.5):
        super(LaplacianGraphOp, self).__init__(prop_steps)
        self.__r = r

    def _construct_adj(self, adj):
        if isinstance(adj, sp.csr_matrix):
            adj = adj.tocoo()
        elif not isinstance(adj, sp.coo_matrix):
            raise TypeError("The adjacency matrix must be a scipy.sparse.coo_matrix/csr_matrix!")

        adj_normalized = adj_to_symmetric_norm(adj, self.__r)
        return adj_normalized.tocsr()



import time
import torch

from sgl.dataset import Dblp
from sgl.models.hetero import NARS_SIGN, Fast_NARS_SGC_WithLearnableWeights
from sgl.tasks import HeteroNodeClassification
from sgl.utils import GpuWithMaxFreeMem

# Hyperparameters
PROP_STEPS = 3
HIDDEN_DIM = 256
NUM_LAYERS = 2
NUM_EPOCHS = 10
LR = 0.01
WEIGHT_DECAY = 0.0
BATCH_SIZE = 10000


def OneTrial(dataset, random_subgraph_num: int, subgraph_edge_type_num: int) -> float:
    predict_class = dataset.TYPE_OF_NODE_TO_PREDICT
    model = Fast_NARS_SGC_WithLearnableWeights(prop_steps=PROP_STEPS,
                                               feat_dim=dataset.data.num_features[predict_class],
                                               output_dim=dataset.data.num_classes[predict_class],
                                               hidden_dim=HIDDEN_DIM, num_layers=NUM_LAYERS,
                                               random_subgraph_num=random_subgraph_num)

    device = torch.device(
        f"cuda:{GpuWithMaxFreeMem()}" if torch.cuda.is_available() else "cpu")
    classification = HeteroNodeClassification(dataset, predict_class, model,
                                              lr=LR, weight_decay=WEIGHT_DECAY,
                                              epochs=NUM_EPOCHS, device=device,
                                              train_batch_size=BATCH_SIZE,
                                              eval_batch_size=BATCH_SIZE,
                                              random_subgraph_num=random_subgraph_num,
                                              subgraph_edge_type_num=subgraph_edge_type_num,
                                              seed=int(time.time()))

    test_acc = classification.test_acc
    raw_weight = classification.subgraph_weight
    weight_sum = raw_weight.sum()
    normalized_weight = raw_weight / weight_sum
    print(normalized_weight)

    return test_acc


def main():
    dataset = Dblp(root='.', path_of_zip='./dataset/DBLP_processed.zip')

    with open('search_result.txt', 'w') as output:
        for random_subgraph_num in range(2, 20):
            for subgraph_edge_type_num in range(2, dataset.edge_type_cnt):
                for iteration in range(3):
                    output.writelines(
                        f'''random_subgraph_num: {random_subgraph_num}, ''' +
                        f'''subgraph_edge_type_num: {subgraph_edge_type_num}, ''' +
                        f'''iteration:{iteration}''' + '\t')
                    test_acc = OneTrial(
                        dataset, random_subgraph_num, subgraph_edge_type_num)
                    output.writelines(f'test accuracy: {test_acc}\n')
                    output.flush()


if __name__ == '__main__':
    main()



import numpy as np
import os.path as osp
import pickle as pkl
import torch

from sgl.data.base_data import Graph
from sgl.data.base_dataset import NodeDataset
from sgl.dataset.utils import pkl_read_file, download_to, random_split_dataset


class Airports(NodeDataset):
    # The number of nodes in training needs to be carefully selected
    def __init__(self, name="usa", root="./", split="official", num_train_per_class=100, num_valid_per_class=20):
        name = name.lower()
        if name not in ["usa", "brazil", "europe"]:
            raise ValueError("Dataset name not found!")
        super(Airports, self).__init__(root + "Airports/", name)

        self._data = pkl_read_file(self.processed_file_paths)
        self._split = split
        self._num_train_per_class = num_train_per_class
        self._num_valid_per_class = num_valid_per_class
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(
            split)

    @property
    def raw_file_paths(self):
        return [
            osp.join(self._raw_dir, f'{self._name}-airports.edgelist'),
            osp.join(self._raw_dir, f'labels-{self.name}-airports.txt')
        ]

    @property
    def processed_file_paths(self):
        return osp.join(self._processed_dir, f"{self._name}.graph")

    def _download(self):
        edge_url = ('https://github.com/leoribeiro/struc2vec/'
                    'raw/master/graph/{}-airports.edgelist')
        label_url = ('https://github.com/leoribeiro/struc2vec/'
                     'raw/master/graph/labels-{}-airports.txt')

        edge_url = edge_url.format(self._name)
        label_url = label_url.format(self._name)

        print(edge_url)
        download_to(edge_url, self.raw_file_paths[0])
        print(label_url)
        download_to(label_url, self.raw_file_paths[1])

    def _process(self):
        index_map, ys = {}, []
        with open(self.raw_file_paths[1], 'r') as f:
            data = f.read().split('\n')[1:-1]
            for i, row in enumerate(data):
                idx, y = row.split()
                index_map[int(idx)] = i
                ys.append(int(y))
        labels = torch.LongTensor(ys)

        # One-hot features
        features = np.eye(labels.shape[0])
        num_node = features.shape[0]
        node_type = "airport"

        edge_indices = []
        with open(self.raw_file_paths[0], 'r') as f:
            data = f.read().split('\n')[:-1]
            for row in data:
                src, dst = row.split()
                edge_indices.append([index_map[int(src)], index_map[int(dst)]])
        edge_index = np.array(edge_indices).T
        row, col = edge_index[0], edge_index[1]

        # Set default edge weight to 1
        edge_weight = np.ones(len(row))
        edge_type = "airport__to__airport"

        g = Graph(row, col, edge_weight, num_node, node_type, edge_type, x=features, y=labels)
        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "official":
            index_map, ys = {}, []
            with open(self.raw_file_paths[1], 'r') as f:
                data = f.read().split('\n')[1:-1]
                for i, row in enumerate(data):
                    idx, y = row.split()
                    index_map[int(idx)] = i
                    ys.append(int(y))
            labels = np.array(ys)

            num_train_per_class = self._num_train_per_class
            num_val = self._num_valid_per_class
            train_idx, val_idx, test_idx = np.empty(0), np.empty(0), np.empty(0)
            for i in range(self.num_classes):
                idx = np.nonzero(labels == i)[0]
                train_idx = np.append(train_idx, idx[:num_train_per_class])
                val_idx = np.append(val_idx, idx[num_train_per_class: num_train_per_class + num_val])
                test_idx = np.append(test_idx, idx[num_train_per_class + num_val:])
            train_idx.reshape(-1)
            val_idx.reshape(-1)
            test_idx.reshape(-1)


        elif split == "random":
            train_idx, val_idx, test_idx = random_split_dataset(self.num_node)
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



import itertools
import numpy as np
import os
import os.path as osp
import torch
import warnings
from scipy.sparse import csr_matrix

from sgl.data.base_data import Node, Edge
from sgl.data.utils import file_exist, to_undirected
from sgl.dataset.choose_edge_type import ChooseMultiSubgraphs


# Base class for node-level tasks
class NodeDataset:
    def __init__(self, root, name):
        self._name = name
        self._root = osp.join(root, name)
        self._raw_dir = osp.join(self._root, "raw")
        self._processed_dir = osp.join(self._root, "processed")
        self._data = None
        self._train_idx, self._val_idx, self._test_idx = None, None, None
        self.__preprocess()

    @property
    def name(self):
        return self._name

    @property
    def raw_file_paths(self):
        raise NotImplementedError

    @property
    def processed_file_paths(self):
        raise NotImplementedError

    def _download(self):
        raise NotImplementedError

    def _process(self):
        raise NotImplementedError

    def __preprocess(self):
        if file_exist(self.raw_file_paths):
            print("Files already downloaded.")
        else:
            print("Downloading...")
            if not file_exist(self._raw_dir):
                os.makedirs(self._raw_dir)
            self._download()
            print("Downloading done!")

        if file_exist(self.processed_file_paths):
            print("Files already processed.")
        else:
            print("Processing...")
            if not file_exist(self._processed_dir):
                os.makedirs(self._processed_dir)
            self._process()
            print("Processing done!")

    @property
    def data(self):
        return self._data

    @property
    def x(self):
        return self._data.x

    @x.setter
    def x(self, x):
        self._data.x = x

    @property
    def y(self):
        return self._data.y

    @y.setter
    def y(self, y):
        self._data.y = y

    @property
    def adj(self):
        return self._data.adj

    @property
    def edge_type(self):
        return self._data.edge_type

    @property
    def edge_type_cnt(self):
        return len(self.edge_types)

    @property
    def node_type(self):
        return self._data.node_type

    @property
    def train_idx(self):
        return self._train_idx

    @property
    def val_idx(self):
        return self._val_idx

    @property
    def test_idx(self):
        return self._test_idx

    @property
    def num_features(self):
        return self._data.num_features

    @property
    def num_classes(self):
        return self._data.num_classes

    @property
    def num_node(self):
        return self._data.num_node


# Base class for graph-level tasks
class GraphDataset:
    pass


# Base class for heterogeneous node-level tasks
class HeteroNodeDataset:
    def __init__(self, root, name):
        self._name = name
        self._root = osp.join(root, name)
        self._raw_dir = osp.join(self._root, "raw")
        self._processed_dir = osp.join(self._root, "processed")
        self._data = None
        self._train_idx, self._val_idx, self._test_idx = None, None, None
        self.__preprocess()

    @property
    def name(self):
        return self._name

    @property
    def raw_file_paths(self):
        raise NotImplementedError

    @property
    def processed_file_paths(self):
        raise NotImplementedError

    def _download(self):
        raise NotImplementedError

    def _process(self):
        raise NotImplementedError

    def __preprocess(self):
        if file_exist(self.raw_file_paths):
            print("Files already downloaded.")
        else:
            print("Downloading...")
            if not file_exist(self._raw_dir):
                os.makedirs(self._raw_dir)
            self._download()
            print("Downloading done!")

        if file_exist(self.processed_file_paths):
            print("Files already processed.")
        else:
            print("Processing...")
            if not file_exist(self._processed_dir):
                os.makedirs(self._processed_dir)
            self._process()
            print("Processing done!")

    def __getitem__(self, key):
        if key in self.data.edge_types:
            return self.data[key]
        elif key in self.data.node_types:
            return self.data[key]
        else:
            raise ValueError("Please input valid edge type or node type!")

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError("Edge type or node type must be a string!")
        if key in self.data.edge_types:
            if not isinstance(value, Edge):
                raise TypeError(
                    "Please organize the dataset using the Edge class!")
            # more restrictions

            self.data.edges_dict[key] = value
        elif key in self.data.node_types:
            if not isinstance(value, Node):
                raise TypeError(
                    "Please organize the dataset using the Node class!")
            # more restrictions

            self.data.nodes_dict[key] = value
        else:
            raise ValueError("Please input valid edge type or node type!")

    @property
    def data(self):
        return self._data

    @property
    def y_dict(self):
        return self._data.y_dict

    @property
    def node_types(self):
        return self._data.node_types

    @property
    def edge_types(self):
        return self._data.edge_types

    @property
    def edge_type_cnt(self):
        return len(self.edge_types)

    @property
    def train_idx(self):
        return self._train_idx

    @property
    def val_idx(self):
        return self._val_idx

    @property
    def test_idx(self):
        return self._test_idx

    # return a sampled adjacency matrix containing edges of given edge types
    def sample_by_edge_type(self, edge_types, undirected=True):
        if not isinstance(edge_types, (str, list, tuple)):
            raise TypeError(
                "The given edge types must be a string or a list or a tuple!")
        elif isinstance(edge_types, str):
            edge_types = [edge_types]
        elif isinstance(edge_types, (list, tuple)):
            for edge_type in edge_types:
                if not isinstance(edge_type, str):
                    raise TypeError("Edge type must be a string!")

        pre_sampled_node_types = []
        for edge_type in edge_types:
            pre_sampled_node_types = pre_sampled_node_types + \
                                     [edge_type.split('__')[0], edge_type.split('__')[2]]
        pre_sampled_node_types = list(set(pre_sampled_node_types))

        sampled_node_types = []
        node_id_offsets = {}
        node_count = 0
        for node_type in self.node_types:
            if node_type in pre_sampled_node_types:
                sampled_node_types.append(node_type)

            node_id_offsets[node_type] = node_count
            node_count = node_count + self._data.num_node[node_type]

        num_node = 0
        feature = None
        node_id = None
        node_id_offset = {}
        for node_type in sampled_node_types:
            node_id_offset[node_type] = node_id_offsets[node_type] - num_node
            num_node = num_node + self._data.num_node[node_type]

            current_feature = torch.from_numpy(self._data[node_type].x)
            if current_feature is None:
                warnings.warn(
                    f'{node_type} nodes have no features!', UserWarning)
            if feature is None:
                feature = current_feature
            else:
                feature = torch.vstack((feature, current_feature))

            if node_id is None:
                node_id = self._data.node_id_dict[node_type][:]
            else:
                node_id = node_id + self._data.node_id_dict[node_type]

        rows, cols = None, None
        for edge_type in edge_types:
            row_temp, col_temp = self._data[edge_type].edge_index

            node_type_of_row = edge_type.split('__')[0]
            node_type_of_col = edge_type.split('__')[2]
            row_temp = row_temp - node_id_offset[node_type_of_row]
            col_temp = col_temp - node_id_offset[node_type_of_col]
            if undirected is True and node_type_of_row != node_type_of_col:
                row_temp, col_temp = to_undirected((row_temp, col_temp))

            if rows is None:
                rows, cols = row_temp, col_temp
            else:
                rows = torch.hstack((rows, row_temp))
                cols = torch.hstack((cols, col_temp))

        edge_weight = torch.ones(len(rows))
        adj = csr_matrix((edge_weight.numpy(), (rows.numpy(),
                                                cols.numpy())), shape=(num_node, num_node))

        # remove previously existed undirected edges
        adj.data = torch.ones(len(adj.data)).numpy()

        return adj, feature.numpy(), torch.LongTensor(node_id)

    # return sampled adjacency matrix containing the given meta-path, "xxx__to__xxx__to...__xxx"
    def sample_by_meta_path(self, meta_path, undirected=True):
        if isinstance(meta_path, str):
            if len(meta_path.split("__")) == 3:
                return self.sample_by_edge_type(meta_path, undirected)

        node_types = meta_path.split("__")

        node_type_st, node_type_ed = node_types[0], node_types[-1]
        sampled_node_types = []
        num_node = 0
        for node_type in self.node_types:
            if node_type in [node_type_st, node_type_ed]:
                sampled_node_types.append(node_type)
                num_node = num_node + self._data.num_node[node_type]

        feature = None
        node_id = None
        for node_type in sampled_node_types:
            current_feature = torch.from_numpy(self._data[node_type].x)
            if current_feature is None:
                warnings.warn(
                    f'{node_type} nodes have no features!', UserWarning)
            if feature is None:
                feature = current_feature
            else:
                feature = torch.vstack((feature, current_feature))

            if node_id is None:
                node_id = self._data.node_id_dict[node_type][:]
            else:
                node_id = node_id + self._data.node_id_dict[node_type]

        # two at a time
        adj = None
        for i in range(int((len(node_types) - 1) / 2)):
            edge_type = "__".join(
                [node_types[i * 2], "to", node_types[(i + 1) * 2]])
            row, col = self._data[edge_type].edge_index
            edge_weight = torch.ones(len(row))
            adj_temp = csr_matrix(
                (edge_weight.numpy(), (row.numpy(), col.numpy())))

            # extremely slow
            if adj is None:
                adj = adj_temp
            else:
                adj = adj * adj_temp

        adj = adj.tocoo()
        row, col, data = torch.LongTensor(adj.row), torch.LongTensor(
            adj.col), torch.FloatTensor(adj.data)

        st_index, ed_index = self.node_types.index(
            node_type_st), self.node_types.index(node_type_ed)
        if st_index == ed_index:
            for node_type in self.node_types[:st_index]:
                row = row - self._data.num_node[node_type]
                col = col - self._data.num_node[node_type]
        else:
            if st_index < ed_index:
                for node_type in self.node_types[:st_index]:
                    row = row - self._data.num_node[node_type]
                    col = col - self._data.num_node[node_type]
                for node_type in self.node_types[st_index:ed_index]:
                    col = col - self._data.num_node[node_type]
                col = col + self._data.num_node[node_type_st]
            else:
                for node_type in self.node_types[:ed_index]:
                    print(node_type)
                    col = col - self._data.num_node[node_type]
                    row = row - self._data.num_node[node_type]
                for node_type in self.node_types[ed_index:st_index]:
                    row = row - self._data.num_node[node_type]
                row = row + self._data.num_node[node_type_ed]

        if undirected is True:
            data = torch.ones(2 * len(data))
            row, col = to_undirected((row, col))
        adj = csr_matrix(
            (data.numpy(), (row.numpy(), col.numpy())), shape=(num_node, num_node))

        # remove existed self loops
        adj.data = torch.ones(len(adj.data)).numpy()
        return adj, feature.numpy(), torch.LongTensor(node_id)

    # return a dict of sub-graphs that contain all the combinations of given edge types and sampled number
    def nars_preprocess(self, edge_types, predict_class, random_subgraph_num, subgraph_edge_type_num):
        if not isinstance(edge_types, (str, list, tuple)):
            raise TypeError(
                "The given edge types must be a string or a list or a tuple!")
        elif isinstance(edge_types, str):
            edge_types = [edge_types]
        elif isinstance(edge_types, (list, tuple)):
            for edge_type in edge_types:
                if not isinstance(edge_type, str):
                    raise TypeError("Edge type must be a string!")

        adopted_edge_type_combinations = ChooseMultiSubgraphs(
            subgraph_num=random_subgraph_num,
            edge_type_num=subgraph_edge_type_num,
            edge_types=edge_types,
            predict_class=predict_class
        )

        if random_subgraph_num > len(adopted_edge_type_combinations):
            random_subgraph_num = len(adopted_edge_type_combinations)
            warnings.warn(
                "The input random_subgraph_num exceeds the number of all the combinations of edge types!"
                f"\nThe random_subgraph_num has been set to {len(adopted_edge_type_combinations)}.", UserWarning)

        chosen_idx = np.random.choice(np.arange(len(adopted_edge_type_combinations)), size=random_subgraph_num,
                                      replace=False)
        chosen_edge_types = [tuple(edge_type) for edge_type in np.array(
            adopted_edge_type_combinations)[chosen_idx]]
        subgraph_dict = {}
        for chosen_edge_type in chosen_edge_types:
            print(chosen_edge_type)
            subgraph_dict[chosen_edge_type] = self.sample_by_edge_type(
                chosen_edge_type)

        return subgraph_dict



from sgl.models.base_model import BaseSGAPModel
from sgl.models.simple_models import ResMultiLayerPerceptron
from sgl.operators.graph_op import PprGraphOp
from sgl.operators.message_op import LearnableWeightedMessageOp
        

class PASCA_V1(BaseSGAPModel):
    def __init__(self, prop_steps, feat_dim, output_dim, hidden_dim, num_layers):
        super(PASCA_V1, self).__init__(prop_steps, feat_dim, output_dim)

        self._pre_graph_op = PprGraphOp(prop_steps, r=0.5, alpha=0.1)
        self._pre_msg_op = LearnableWeightedMessageOp(1, prop_steps + 1, "simple", feat_dim)
        self._base_model = ResMultiLayerPerceptron(feat_dim, hidden_dim, num_layers, output_dim, 0.8)



import numpy as np
import platform
import scipy.sparse as sp
import torch
import torch.nn as nn
from torch import Tensor

from sgl.operators.utils import csr_sparse_dense_matmul, cuda_csr_sparse_dense_matmul


class GraphOp:
    def __init__(self, prop_steps):
        self._prop_steps = prop_steps
        self._adj = None

    def _construct_adj(self, adj):
        raise NotImplementedError

    def propagate(self, adj, feature):
        self._adj = self._construct_adj(adj)

        if not isinstance(adj, sp.csr_matrix):
            raise TypeError("The adjacency matrix must be a scipy csr sparse matrix!")
        elif not isinstance(feature, np.ndarray):
            raise TypeError("The feature matrix must be a numpy.ndarray!")
        elif self._adj.shape[1] != feature.shape[0]:
            raise ValueError("Dimension mismatch detected for the adjacency and the feature matrix!")

        prop_feat_list = [feature]
        for _ in range(self._prop_steps):
            if platform.system() == "Linux":
                feat_temp = csr_sparse_dense_matmul(self._adj, prop_feat_list[-1])
            else:
                feat_temp = self._adj.dot(prop_feat_list[-1])
            prop_feat_list.append(feat_temp)
        return [torch.FloatTensor(feat) for feat in prop_feat_list]


# Might include training parameters
class MessageOp(nn.Module):
    def __init__(self, start=None, end=None):
        super(MessageOp, self).__init__()
        self._aggr_type = None
        self._start, self._end = start, end

    @property
    def aggr_type(self):
        return self._aggr_type

    def _combine(self, feat_list):
        return NotImplementedError

    def aggregate(self, feat_list):
        if not isinstance(feat_list, list):
            return TypeError("The input must be a list consists of feature matrices!")
        for feat in feat_list:
            if not isinstance(feat, Tensor):
                raise TypeError("The feature matrices must be tensors!")

        return self._combine(feat_list)



import math
import random
import warnings
from typing import List, Tuple

EDGE_TYPE_DELIMITER = '__to__'


# Without sorting. Keep the original order.
def EdgeTypeStr2Tuple(edge_type: str) -> Tuple[str]:
    edge_type_list = edge_type.split(EDGE_TYPE_DELIMITER)
    return tuple(edge_type_list)


# Due to the existence of directed edges in both directions,
# e.g., paper__to__author and author__to__paper may exist simultaneously.
def RemoveDuplicateEdgeType(edge_types: List) -> List[str]:
    unique_edge_types = []
    for et in edge_types:
        et_tuple = EdgeTypeStr2Tuple(et)
        reversed_et = et_tuple[1] + EDGE_TYPE_DELIMITER + et_tuple[0]
        if reversed_et not in unique_edge_types:
            unique_edge_types.append(et)
    return unique_edge_types


# There must be no duplicate elements in edge_types.
def ChooseEdgeType(edge_type_num: int, edge_types: List, predict_class: str) -> Tuple[str]:
    # Chosen edge types should be connected to the predict_class.
    # In other words, they should interset with explored_node_type_set.
    explored_node_type_set = {predict_class}
    chosen_edge_types_list = []
    candidate_edge_types_list = []
    # Due to the lack of connection,
    # these edge types cannot be chosen at the moment.
    other_edge_types_set = set(edge_types)

    for _ in range(edge_type_num):
        # Move edge types from other_edge_types_set to candidate_edge_types_list.
        edge_types_to_move = [
            et for et in other_edge_types_set
            if len(set(EdgeTypeStr2Tuple(et)) &
                   explored_node_type_set) > 0]
        candidate_edge_types_list += edge_types_to_move
        other_edge_types_set -= set(edge_types_to_move)

        if len(candidate_edge_types_list) == 0:
            warnings.warn(
                f"Can't find enough ({edge_type_num}) edge types!", UserWarning)
            break
        # Move edge types from candidate_edge_types_list to chosen_edge_types_list.
        new_edge_type = random.choice(candidate_edge_types_list)
        chosen_edge_types_list.append(new_edge_type)
        candidate_edge_types_list.remove(new_edge_type)
        explored_node_type_set |= set(EdgeTypeStr2Tuple(new_edge_type))

    return tuple(sorted(chosen_edge_types_list))


# Ways of choosing k items from n items.
def Combination(n: int, k: int) -> int:
    if n < 0 or k < 0:
        raise ValueError('n < 0 or k < 0!')
    result = 1
    for i in range(k):
        result = result * (n - i) // (i + 1)
    return result


def ChooseMultiSubgraphs(subgraph_num: int, edge_type_num: int,
                         edge_types: List, predict_class: str) -> List[Tuple[str]]:
    subgraph_edge_types_list = []
    unique_edge_type = RemoveDuplicateEdgeType(edge_types)
    if edge_type_num > len(unique_edge_type):
        return subgraph_edge_types_list

    # Estimate by "coupon collector"
    maximal_reasonable_steps = 10 * Combination(len(unique_edge_type), edge_type_num) * \
                               (math.log2(Combination(len(unique_edge_type), edge_type_num)) + 1)
    step_cnt = 0

    for _ in range(subgraph_num):
        while True:
            # Avoid infinite loop.
            step_cnt += 1
            if step_cnt > maximal_reasonable_steps:
                warnings.warn(
                    f"Can't find enough ({subgraph_num}) subgraphs!", UserWarning)
                break

            new_subgraph_edge_types = ChooseEdgeType(
                edge_type_num, unique_edge_type, predict_class)
            if new_subgraph_edge_types in subgraph_edge_types_list:
                continue
            if len(new_subgraph_edge_types) > 0:
                subgraph_edge_types_list.append(new_subgraph_edge_types)
            break
    return subgraph_edge_types_list


def main():
    edge_types = ['paper__to__author', 'paper__to__paper', 'paper__to__venue',
                  'author__to__paper', 'author__to__author', 'author__to__venue',
                  'venue__to__paper', 'venue__to__author', 'venue__to__venue',
                  'paper__to__keyword', 'keyword__to__paper', 'keyword__to__keyword']
    predict_class = 'paper'
    subgraph_num = 5
    edge_type_num = 5
    subgraph_edge_types_list = ChooseMultiSubgraphs(
        subgraph_num, edge_type_num, edge_types, predict_class)
    for ele in subgraph_edge_types_list:
        print(ele)


if __name__ == '__main__':
    main()



from .acm import Acm
from .actor import Actor
from .airports import Airports
from .amazon import Amazon
from .amazon_product import AmazonProduct
from .coauthor import Coauthor
from .dblp import Dblp
from .facebook import Facebook
from .flickr import Flickr
from .github import Github
from .karateclub import KarateClub
from .linkx_dataset import LINKXDataset
from .nell import Nell
from .ogbn import Ogbn
from .ogbn_mag import OgbnMag
from .planetoid import Planetoid
from .reddit import Reddit
from .twitch import Twitch
from .webkb import WebKB
from .wikics import Wikics

__all__ = [
    "Acm",
    "Actor",
    "Airports",
    "AmazonProduct",
    "Amazon",
    "Coauthor",
    "Dblp",
    "Facebook",
    "Flickr",
    "Github",
    "KarateClub",
    "LINKXDataset",
    "Nell",
    "OgbnMag",
    "Ogbn",
    "Planetoid",
    "Reddit",
    "Twitch",
    "WebKB",
    "Wikics",
]



import matplotlib
import numpy as np
from munkres import Munkres
from sklearn import metrics
from sklearn.manifold import TSNE

matplotlib.use('Agg')
import matplotlib.pyplot as plt


class clustering_metrics:
    def __init__(self, true_label, predict_label):
        self.true_label = true_label
        self.pred_label = predict_label

    def clusteringAcc(self):
        # best mapping between true_label and predict label
        l1 = list(set(self.true_label))
        numclass1 = len(l1)

        l2 = list(set(self.pred_label))
        numclass2 = len(l2)

        if numclass1 != numclass2:
            #            print('Class Not equal, Error!!!!')
            return 0

        cost = np.zeros((numclass1, numclass2), dtype=int)
        for i, c1 in enumerate(l1):
            mps = [i1 for i1, e1 in enumerate(self.true_label) if e1 == c1]
            for j, c2 in enumerate(l2):
                mps_d = [i1 for i1 in mps if self.pred_label[i1] == c2]

                cost[i][j] = len(mps_d)

        # match two clustering results by Munkres algorithm
        m = Munkres()
        cost = cost.__neg__().tolist()

        indexes = m.compute(cost)

        # get the match results
        new_predict = np.zeros(len(self.pred_label))
        for i, c in enumerate(l1):
            # correponding label in l2:
            c2 = l2[indexes[i][1]]

            # ai is the index with label==c2 in the pred_label list
            ai = [ind for ind, elm in enumerate(self.pred_label) if elm == c2]
            new_predict[ai] = c
        acc = metrics.accuracy_score(self.true_label, new_predict)
        f1_macro = metrics.f1_score(self.true_label, new_predict, average='macro')
        precision_macro = metrics.precision_score(self.true_label, new_predict, average='macro')
        recall_macro = metrics.recall_score(self.true_label, new_predict, average='macro')
        f1_micro = metrics.f1_score(self.true_label, new_predict, average='micro')
        precision_micro = metrics.precision_score(self.true_label, new_predict, average='micro')
        recall_micro = metrics.recall_score(self.true_label, new_predict, average='micro')

        return acc, f1_macro, precision_macro, recall_macro, f1_micro, precision_micro, recall_micro

    def evaluationClusterModelFromLabel(self):
        nmi = metrics.normalized_mutual_info_score(self.true_label, self.pred_label)
        adjscore = metrics.adjusted_rand_score(self.true_label, self.pred_label)
        acc, f1_macro, precision_macro, recall_macro, f1_micro, precision_micro, recall_micro = self.clusteringAcc()

        return acc, nmi, adjscore

    @staticmethod
    def plot(X, fig, col, size, true_labels):
        ax = fig.add_subplot(1, 1, 1)
        for i, point in enumerate(X):
            ax.scatter(point[0], point[1], lw=0, s=size, c=col[true_labels[i]])

    def plotClusters(self, tqdm, hidden_emb, true_labels):
        tqdm.write('Start plotting using TSNE...')
        # Doing dimensionality reduction for plotting
        tsne = TSNE(n_components=2)
        X_tsne = tsne.fit_transform(hidden_emb)
        # Plot figure
        fig = plt.figure()
        self.plot(X_tsne, fig, ['red', 'green', 'blue', 'brown', 'purple', 'yellow', 'pink', 'orange'], 40, true_labels)
        plt.axis("off")
        fig.savefig("plot.png", dpi=120)
        tqdm.write("Finished plotting")



from sgl.models.base_model import BaseSGAPModel
from sgl.models.simple_models import LogisticRegression, ResMultiLayerPerceptron
from sgl.operators.graph_op import LaplacianGraphOp, PprGraphOp
from sgl.operators.message_op import LastMessageOp, ConcatMessageOp, MeanMessageOp, SimpleWeightedMessageOp, \
    LearnableWeightedMessageOp, IterateLearnableWeightedMessageOp, SumMessageOp, MaxMessageOp, MinMessageOp


class SearchModel(BaseSGAPModel):
    def __init__(self, arch, feat_dim, output_dim, hidden_dim):
        prop_steps = arch[0]
        prop_types = arch[1]
        mesg_types = arch[2]
        num_layers = arch[3]
        post_steps = arch[4]
        post_types = arch[5]
        pmsg_types = arch[6]
        super(SearchModel, self).__init__(prop_steps, feat_dim, output_dim)

        if prop_types == 1:
            self._pre_graph_op = LaplacianGraphOp(prop_steps, r=0.5)
        elif prop_types == 2:
            self._pre_graph_op = PprGraphOp(prop_steps, r=0.5, alpha=0.1)
        elif prop_types == 3:
            self._pre_graph_op = PprGraphOp(prop_steps, r=0.5, alpha=0.2)
        elif prop_types == 4:
            self._pre_graph_op = PprGraphOp(prop_steps, r=0.5, alpha=0.3)

        if mesg_types == 0:
            self._pre_msg_op = LastMessageOp()
        elif mesg_types == 1:
            self._pre_msg_op = ConcatMessageOp(start=0, end=prop_steps + 1)
            feat_dim *= prop_steps + 1
        elif mesg_types == 2:
            self._pre_msg_op = MeanMessageOp(start=0, end=prop_steps + 1)
        elif mesg_types == 3:
            self._pre_msg_op = SumMessageOp(start=0, end=prop_steps + 1)
        elif mesg_types == 4:
            self._pre_msg_op = MaxMessageOp(start=0, end=prop_steps + 1)
        elif mesg_types == 5:
            self._pre_msg_op = MinMessageOp(start=0, end=prop_steps + 1)
        elif mesg_types == 6:
            self._pre_msg_op = SimpleWeightedMessageOp(0, prop_steps + 1, "alpha", 0.85)
        elif mesg_types == 7:
            self._pre_msg_op = LearnableWeightedMessageOp(1, prop_steps + 1, "gate", feat_dim)
        elif mesg_types == 8:
            self._pre_msg_op = LearnableWeightedMessageOp(1, prop_steps + 1, "simple", feat_dim)

        if num_layers == 1:
            self._base_model = LogisticRegression(feat_dim, output_dim)
        else:
            self._base_model = ResMultiLayerPerceptron(feat_dim, hidden_dim, num_layers, output_dim)

        if post_types != 0 and post_steps != 0:
            if post_types == 1:
                self._post_graph_op = LaplacianGraphOp(post_steps, r=0.5)
            elif post_types == 2:
                self._post_graph_op = PprGraphOp(post_steps, r=0.5, alpha=0.1)
            elif post_types == 3:
                self._post_graph_op = PprGraphOp(post_steps, r=0.5, alpha=0.2)
            elif post_types == 4:
                self._post_graph_op = PprGraphOp(post_steps, r=0.5, alpha=0.3)

            if pmsg_types == 0:
                self._post_msg_op = LastMessageOp()
            elif pmsg_types == 1:
                self._post_msg_op = MeanMessageOp(start=0, end=post_steps + 1)
            elif pmsg_types == 2:
                self._post_msg_op = SumMessageOp(start=0, end=post_steps + 1)
            elif pmsg_types == 3:
                self._post_msg_op = MaxMessageOp(start=0, end=post_steps + 1)
            elif pmsg_types == 4:
                self._post_msg_op = MinMessageOp(start=0, end=post_steps + 1)
            elif pmsg_types == 5:
                self._post_msg_op = SimpleWeightedMessageOp(0, post_steps + 1, "alpha", 0.85)



from sgl.models.base_model import BaseSGAPModel
from sgl.models.simple_models import LogisticRegression, MultiLayerPerceptron, ResMultiLayerPerceptron
from sgl.operators.graph_op import LaplacianGraphOp, PprGraphOp
from sgl.operators.message_op import LastMessageOp, ConcatMessageOp, MeanMessageOp, SimpleWeightedMessageOp, \
    LearnableWeightedMessageOp, IterateLearnableWeightedMessageOp, SumMessageOp, MaxMessageOp, MinMessageOp


class SearchModelDist(BaseSGAPModel):
    def __init__(self, arch, feat_dim, output_dim, hidden_dim):
        prop_steps = arch[0]
        prop_types = arch[1]
        mesg_types = arch[2]
        num_layers = arch[3]
        post_steps = arch[4]
        post_types = arch[5]
        pmsg_types = arch[6]
        super(SearchModelDist, self).__init__(prop_steps, feat_dim, output_dim)
        
        if prop_types == 1:
            self._pre_graph_op = LaplacianGraphOp(prop_steps, r=0.5)
        elif prop_types == 2:
            self._pre_graph_op = PprGraphOp(prop_steps, r=0.5, alpha=0.1)
        elif prop_types == 3:
            self._pre_graph_op = PprGraphOp(prop_steps, r=0.5, alpha=0.2)
        elif prop_types == 4:
            self._pre_graph_op = PprGraphOp(prop_steps, r=0.5, alpha=0.3)        

        if mesg_types == 0:
            self._pre_msg_op = LastMessageOp()
        elif mesg_types == 1:
            self._pre_msg_op = ConcatMessageOp(start=0, end=prop_steps + 1)
            feat_dim *= prop_steps + 1 
        elif mesg_types == 2:
            self._pre_msg_op = MeanMessageOp(start=0, end=prop_steps + 1)
        elif mesg_types == 3:
            self._pre_msg_op = SumMessageOp(start=0, end=prop_steps + 1)
        elif mesg_types == 4:
            self._pre_msg_op = MaxMessageOp(start=0, end=prop_steps + 1)
        elif mesg_types == 5:
            self._pre_msg_op = MinMessageOp(start=0, end=prop_steps + 1)
        elif mesg_types == 6:
            self._pre_msg_op = SimpleWeightedMessageOp(0, prop_steps + 1, "alpha", 0.85)
        elif mesg_types == 7:
            self._pre_msg_op = LearnableWeightedMessageOp(0, prop_steps, "gate", feat_dim)
        elif mesg_types == 8:
            self._pre_msg_op = IterateLearnableWeightedMessageOp(0, prop_steps + 1, "recursive", feat_dim)
        

        if num_layers == 1:
            self._base_model = LogisticRegression(feat_dim, output_dim)
        elif num_layers == 2:
            self._base_model = MultiLayerPerceptron(feat_dim, hidden_dim, num_layers, output_dim)
        else:
            self._base_model = ResMultiLayerPerceptron(feat_dim, hidden_dim, num_layers, output_dim)

        if post_types != 0 and post_steps != 0:
            if post_types == 1:
                self._post_graph_op = LaplacianGraphOp(post_steps, r=0.5)
            elif post_types == 2:
                self._post_graph_op = PprGraphOp(post_steps, r=0.5, alpha=0.1)
            elif post_types == 3:
                self._post_graph_op = PprGraphOp(post_steps, r=0.5, alpha=0.2)
            elif post_types == 4:
                self._post_graph_op = PprGraphOp(post_steps, r=0.5, alpha=0.3)
            
            if pmsg_types == 0:
                self._post_msg_op = LastMessageOp()
            elif pmsg_types == 1:
                self._post_msg_op = MeanMessageOp(start=0, end=post_steps + 1)
            elif pmsg_types == 2:
                self._post_msg_op = SumMessageOp(start=0, end=post_steps + 1)
            elif pmsg_types == 3:
                self._post_msg_op = MaxMessageOp(start=0, end=post_steps + 1)
            elif pmsg_types == 4:
                self._post_msg_op = MinMessageOp(start=0, end=post_steps + 1)
            elif pmsg_types == 5:
                self._post_msg_op = SimpleWeightedMessageOp(0, post_steps + 1, "alpha", 0.85)

    def preprocess(self, adj, feature):
        if self._pre_graph_op is not None:
            self._processed_feat_list = self._pre_graph_op.propagate(
                adj, feature)
        else:
            self._processed_feat_list = [feature]

    # a wrapper of the forward function
    def model_forward(self, idx, device):
        return self.evaluate_forward(idx, device)

    def evaluate_forward(self, idx, device):   
        transferred_feat_list = [feat[idx].to(
                device) for feat in self._processed_feat_list]
        processed_feature = self._pre_msg_op.aggregate(
            transferred_feat_list)

        output = self._base_model(processed_feature)
        return output
            
    def forward(self, transferred_feat_list):   
        processed_feature = self._pre_msg_op.aggregate(
            transferred_feat_list)
            
        output = self._base_model(processed_feature)
        return output



import torch

from sgl.operators.base_op import MessageOp


class MaxMessageOp(MessageOp):
    def __init__(self, start, end):
        super(MaxMessageOp, self).__init__(start, end)
        self._aggr_type = "max"

    def _combine(self, feat_list):
        return torch.stack(feat_list[self._start:self._end], dim=0).max(dim=0)[0]



import numpy as np

from sgl.dataset import Dblp
from sgl.etc.auto_select_edge_type_for_nars import GenerateSubgraphList, OneTrialWithSubgraphList

SUBGRAPH_COMBINATIONS = 20
NUM_REPEAT = 5
SUBGRAPH_CONFIG = [(3, 2), (3, 3)]
NUM_EPOCHS_TO_FIND_WEIGHT = 20


def main():
    dataset = Dblp(root='.', path_of_zip='./dataset/DBLP_processed.zip')
    with open('subgraph_weight.txt', 'w') as output:
        for i in range(SUBGRAPH_COMBINATIONS):
            output.write(f'Subgraph combination {i + 1}\n')
            subgraph_list = GenerateSubgraphList(dataset, SUBGRAPH_CONFIG)
            for j in range(NUM_REPEAT):
                output.write(f'\tIteration {j + 1}\n')
                test_acc, _, subgraph_weight = OneTrialWithSubgraphList(
                    dataset, subgraph_list, NUM_EPOCHS_TO_FIND_WEIGHT)
                index_sort = np.argsort(subgraph_weight.detach().numpy())
                output.write(f'\t\tTest accuracy: {test_acc}\n')
                output.write(f'\t\tSubgraph weight: {index_sort}\n\n')
                output.flush()


if __name__ == '__main__':
    main()
